@@ -54,6 +54,9 @@ default_ignored_tests = {
     "net",
 }
 
+# Some tests are ignored on PG earlier than 17 due to broken MergeAppend cost model there.
+ignored_before_pg17 = default_ignored_tests | {"merge_append_partially_compressed"}
+
 # Tests that we do not run as part of a Flake tests
 flaky_exclude_tests = {
     # Not executed as a flake test since it easily exhausts available
@@ -90,7 +93,8 @@ def build_debug_config(overrides):
             "name": "Debug",
             "os": "ubuntu-22.04",
             "pg_extra_args": "--enable-debug --enable-cassert --with-llvm LLVM_CONFIG=llvm-config-14",
-            "pg_extensions": "postgres_fdw test_decoding pageinspect pgstattuple",
+            "pg_extensions": "postgres_fdw test_decoding",
+            "installcheck": True,
             "pginstallcheck": True,
             "tsdb_build_args": "-DWARNINGS_AS_ERRORS=ON -DREQUIRE_ALL_TESTS=ON",
         }
@@ -158,7 +162,7 @@ def macos_config(overrides):
             "ignored_tests": default_ignored_tests.union(macos_ignored_tests),
             "os": "macos-13",
             "pg_extra_args": "--enable-debug --with-libraries=/usr/local/opt/openssl@3/lib --with-includes=/usr/local/opt/openssl@3/include --without-icu",
-            "pg_extensions": "postgres_fdw test_decoding pageinspect pgstattuple",
+            "pg_extensions": "postgres_fdw test_decoding",
             "pginstallcheck": True,
             "tsdb_build_args": "-DASSERTIONS=ON -DREQUIRE_ALL_TESTS=ON -DOPENSSL_ROOT_DIR=/usr/local/opt/openssl@3",
         }
@@ -168,9 +172,13 @@ def macos_config(overrides):
 
 
 # always test debug build on latest of all supported pg versions
-m["include"].append(build_debug_config({"pg": PG15_LATEST}))
+m["include"].append(
+    build_debug_config({"pg": PG15_LATEST, "ignored_tests": ignored_before_pg17})
+)
 
-m["include"].append(build_debug_config({"pg": PG16_LATEST}))
+m["include"].append(
+    build_debug_config({"pg": PG16_LATEST, "ignored_tests": ignored_before_pg17})
+)
 
 m["include"].append(build_debug_config({"pg": PG17_LATEST}))
 
@@ -205,30 +213,60 @@ m["include"].append(
     )
 )
 
+# test building against PG18beta3
+m["include"].append(
+    build_debug_config(
+        {
+            "pg": "18beta3",
+            "pg_extra_args": "--enable-debug --enable-cassert --without-llvm",
+            "tsdb_build_args": "-DEXPERIMENTAL=ON -DWARNINGS_AS_ERRORS=OFF",
+            "installcheck": False,
+            "pginstallcheck": False,
+            "coverage": False,
+        }
+    )
+)
+
 # if this is not a pull request e.g. a scheduled run or a push
 # to a specific branch like prerelease_test we add additional
 # entries to the matrix
 if not pull_request:
     # add debug test for first supported PG15 version
-    m["include"].append(build_debug_config({"pg": PG15_EARLIEST}))
+    m["include"].append(
+        build_debug_config({"pg": PG15_EARLIEST, "ignored_tests": ignored_before_pg17})
+    )
 
     # add debug test for first supported PG16 version
-    m["include"].append(build_debug_config({"pg": PG16_EARLIEST}))
+    m["include"].append(
+        build_debug_config({"pg": PG16_EARLIEST, "ignored_tests": ignored_before_pg17})
+    )
 
     # add debug test for first supported PG17 version
     if PG17_EARLIEST != PG17_LATEST:
         m["include"].append(build_debug_config({"pg": PG17_EARLIEST}))
 
     # add debug tests for timescaledb on latest postgres release in MacOS
-    m["include"].append(build_debug_config(macos_config({"pg": PG15_LATEST})))
+    m["include"].append(
+        build_debug_config(
+            macos_config({"pg": PG15_LATEST, "ignored_tests": ignored_before_pg17})
+        )
+    )
 
-    m["include"].append(build_debug_config(macos_config({"pg": PG16_LATEST})))
+    m["include"].append(
+        build_debug_config(
+            macos_config({"pg": PG16_LATEST, "ignored_tests": ignored_before_pg17})
+        )
+    )
 
     m["include"].append(build_debug_config(macos_config({"pg": PG17_LATEST})))
 
     # add release test for latest pg releases
-    m["include"].append(build_release_config({"pg": PG15_LATEST}))
-    m["include"].append(build_release_config({"pg": PG16_LATEST}))
+    m["include"].append(
+        build_release_config({"pg": PG15_LATEST, "ignored_tests": ignored_before_pg17})
+    )
+    m["include"].append(
+        build_release_config({"pg": PG16_LATEST, "ignored_tests": ignored_before_pg17})
+    )
     m["include"].append(build_release_config({"pg": PG17_LATEST}))
 
     # add apache only test for latest pg versions
@@ -241,6 +279,7 @@ if not pull_request:
         build_debug_config(
             {
                 "pg": 15,
+                "ignored_tests": ignored_before_pg17,
                 "snapshot": "snapshot",
             }
         )
@@ -249,6 +288,7 @@ if not pull_request:
         build_debug_config(
             {
                 "pg": 16,
+                "ignored_tests": ignored_before_pg17,
                 "snapshot": "snapshot",
             }
         )
