@@ -137,7 +137,7 @@ policy_refresh_cagg_get_refresh_end(const Dimension *dim, const Jsonb *config, b
 	int64 res = get_time_from_config(dim, config, POL_REFRESH_CONF_KEY_END_OFFSET, end_isnull);
 
 	if (*end_isnull)
-		return ts_time_get_end_or_max(ts_dimension_get_partition_type(dim));
+		return ts_time_get_noend_or_max(ts_dimension_get_partition_type(dim));
 	return res;
 }
 
@@ -448,7 +448,7 @@ validate_window_size(const ContinuousAgg *cagg, const CaggPolicyConfig *config)
 
 static void
 parse_offset_arg(const ContinuousAgg *cagg, Oid offset_type, NullableDatum arg,
-				 CaggPolicyOffset *offset)
+				 ContinuousAggPolicyOffset *offset)
 {
 	offset->isnull = arg.isnull;
 
@@ -559,11 +559,7 @@ policy_refresh_cagg_check_for_overlaps(ContinuousAgg *cagg, Jsonb *policy_config
 	if (typcache == NULL || typcache->rngelemtype == NULL)
 		elog(ERROR, "cache lookup failed");
 
-#if PG16_LT
-	RangeType *range = make_range(typcache, &lower, &upper, false);
-#else
-	RangeType *range = make_range(typcache, &lower, &upper, false, NULL);
-#endif
+	RangeType *range = make_range_compat(typcache, &lower, &upper, false, NULL);
 
 	ListCell *lc;
 
@@ -606,11 +602,7 @@ policy_refresh_cagg_check_for_overlaps(ContinuousAgg *cagg, Jsonb *policy_config
 			.lower = false,
 		};
 
-#if PG16_LT
-		RangeType *range_job = make_range(typcache, &lower_job, &upper_job, false);
-#else
-		RangeType *range_job = make_range(typcache, &lower_job, &upper_job, false, NULL);
-#endif
+		RangeType *range_job = make_range_compat(typcache, &lower_job, &upper_job, false, NULL);
 
 		elog(DEBUG1,
 			 "start_offset_job: " INT64_FORMAT ", end_offset_job: " INT64_FORMAT,
