@@ -44,10 +44,6 @@
 #include "debug_point.h"
 #include "error_utils.h"
 #include "errors.h"
-#include "extension_constants.h"
-#include "guc.h"
-#include "hypercore/hypercore_handler.h"
-#include "hypercore/utils.h"
 #include "hypercube.h"
 #include "hypertable.h"
 #include "hypertable_cache.h"
@@ -455,19 +451,6 @@ compress_chunk_impl(Oid hypertable_relid, Oid chunk_relid)
 						NameStr(compress_ht_chunk->fd.schema_name),
 						NameStr(compress_ht_chunk->fd.table_name))));
 
-		/* Since a new compressed relation was created it is necessary to
-		 * invalidate the relcache entry for the chunk because Hypercore TAM
-		 * caches information about the compressed relation in the
-		 * relcache. */
-		if (ts_is_hypercore_am(cxt.srcht_chunk->amoid))
-		{
-			/* Tell other backends */
-			CacheInvalidateRelcacheByRelid(cxt.srcht_chunk->table_id);
-
-			/* Immediately invalidate our own cache */
-			RelationCacheInvalidateEntry(cxt.srcht_chunk->table_id);
-		}
-
 		EventTriggerAlterTableEnd();
 	}
 	else
@@ -756,164 +739,19 @@ tsl_create_compressed_chunk(PG_FUNCTION_ARGS)
 	PG_RETURN_OID(chunk_relid);
 }
 
-static Oid
-set_access_method(Oid relid, const char *amname)
-{
-	AlterTableCmd cmd = {
-		.type = T_AlterTableCmd,
-		.subtype = AT_SetAccessMethod,
-		.name = pstrdup(amname),
-	};
-	bool to_hypercore = strcmp(amname, TS_HYPERCORE_TAM_NAME) == 0;
-	Oid amoid = ts_get_rel_am(relid);
-	Oid new_amoid = get_am_oid(amname, false);
-
-	/* Setting the same access method is a no-op */
-	if (amoid == new_amoid)
-		return relid;
-
-	hypercore_alter_access_method_begin(relid, !to_hypercore);
-	AlterTableInternal(relid, list_make1(&cmd), false);
-
-#if (PG_VERSION_NUM < 150004)
-
-	/*
-	 * Also do a runtime check in order to be ABI compatible with PG server
-	 * upgrades, e.g., upgrading from 15.3 to 15.4 without updating the
-	 * extension.
-	 */
-	const char *version_num_str = GetConfigOption("server_version_num", false, false);
-	int server_version_num;
-
-	if (parse_int(version_num_str, &server_version_num, 0, NULL) && server_version_num < 150004)
-	{
-		/* Fix for PostgreSQL bug where pg_depend was not updated to reflect the
-		 * new dependency between AM and relation. See related PG fix here:
-		 * https://github.com/postgres/postgres/commit/97d89101045fac8cb36f4ef6c08526ea0841a596 */
-		if (changeDependencyFor(RelationRelationId,
-								relid,
-								AccessMethodRelationId,
-								amoid,
-								new_amoid) != 1)
-			elog(ERROR,
-				 "could not change access method dependency for relation \"%s.%s\"",
-				 get_namespace_name(get_rel_namespace(relid)),
-				 get_rel_name(relid));
-	}
-#endif
-	hypercore_alter_access_method_finish(relid, !to_hypercore);
-
-	return relid;
-}
-
-/*
- * When using compress_chunk() with hypercore, there are three cases to
- * handle:
- *
- * 1. Convert from (uncompressed) heap to hypercore
- *
- * 2. Convert from compressed heap to hypercore
- *
- * 3. Recompress a hypercore
- */
-static Oid
-compress_hypercore(Chunk *chunk, bool rel_is_hypercore, UseAccessMethod useam,
-				   bool if_not_compressed, bool recompress)
-{
-	Oid relid = InvalidOid;
-
-	/* Either the chunk is already a hypercore (and in that case recompress),
-	 * or it is being converted to one */
-	Assert(rel_is_hypercore || useam == USE_AM_TRUE);
-
-	if (ts_chunk_is_compressed(chunk) && !rel_is_hypercore)
-	{
-		Assert(useam == USE_AM_TRUE);
-		char *relname = get_rel_name(chunk->table_id);
-		char *relschema = get_namespace_name(get_rel_namespace(chunk->table_id));
-		const RangeVar *rv = makeRangeVar(relschema, relname, -1);
-		/* Do quick migration to hypercore of already compressed data by
-		 * simply changing the access method to hypercore in pg_am. */
-		hypercore_set_am(rv, TS_HYPERCORE_TAM_NAME);
-		hypercore_set_compressed_autovacuum_reloption(chunk, false);
-		return chunk->table_id;
-	}
-
-	switch (useam)
-	{
-		case USE_AM_FALSE:
-			/* Converting from Hypercore to "regular" compressed is currently
-			 * not supported */
-			Assert(rel_is_hypercore);
-			ereport(ERROR,
-					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-					 errmsg("cannot converting to columnstore \"%s\" without using Hypercore "
-							"access method",
-							get_rel_name(chunk->table_id)),
-					 errhint("Convert to rowstore first and then convert to columnstore without "
-							 "Hypercore access method.")));
-			break;
-		case USE_AM_NULL:
-			Assert(rel_is_hypercore);
-			/* Don't forward the truncate to the compressed data during recompression */
-			bool truncate_compressed = hypercore_set_truncate_compressed(false);
-			relid = tsl_compress_chunk_wrapper(chunk, if_not_compressed, recompress);
-			hypercore_set_truncate_compressed(truncate_compressed);
-			break;
-		case USE_AM_TRUE:
-			if (rel_is_hypercore)
-			{
-				/* Don't forward the truncate to the compressed data during recompression */
-				bool truncate_compressed = hypercore_set_truncate_compressed(false);
-				relid = tsl_compress_chunk_wrapper(chunk, if_not_compressed, recompress);
-				hypercore_set_truncate_compressed(truncate_compressed);
-			}
-			else
-			{
-				/* Convert to a compressed hypercore by simply calling ALTER TABLE
-				 * <chunk> SET ACCESS METHOD hypercore */
-				set_access_method(chunk->table_id, TS_HYPERCORE_TAM_NAME);
-				relid = chunk->table_id;
-			}
-			break;
-	}
-
-	return relid;
-}
-
-/*
- * Check the value of the use_access_method argument and use the default value
- * otherwise.
- */
-static UseAccessMethod
-check_useam(UseAccessMethod arg, bool is_hypercore)
-{
-	if (arg == USE_AM_NULL)
-		return is_hypercore ? USE_AM_TRUE :
-							  (UseAccessMethod) ts_guc_default_hypercore_use_access_method;
-	return arg;
-}
-
 Datum
 tsl_compress_chunk(PG_FUNCTION_ARGS)
 {
 	Oid uncompressed_chunk_id = PG_ARGISNULL(0) ? InvalidOid : PG_GETARG_OID(0);
 	bool if_not_compressed = PG_ARGISNULL(1) ? true : PG_GETARG_BOOL(1);
 	bool recompress = PG_ARGISNULL(2) ? false : PG_GETARG_BOOL(2);
-	UseAccessMethod useam;
 
 	ts_feature_flag_check(FEATURE_HYPERTABLE_COMPRESSION);
 
 	TS_PREVENT_FUNC_IF_READ_ONLY();
 	Chunk *chunk = ts_chunk_get_by_relid(uncompressed_chunk_id, true);
-	bool rel_is_hypercore = get_table_am_oid(TS_HYPERCORE_TAM_NAME, false) == chunk->amoid;
-	useam = check_useam(PG_ARGISNULL(3) ? USE_AM_NULL : PG_GETARG_BOOL(3), rel_is_hypercore);
 
-	if (rel_is_hypercore || useam == USE_AM_TRUE)
-		uncompressed_chunk_id =
-			compress_hypercore(chunk, rel_is_hypercore, useam, if_not_compressed, recompress);
-	else
-		uncompressed_chunk_id = tsl_compress_chunk_wrapper(chunk, if_not_compressed, recompress);
+	uncompressed_chunk_id = tsl_compress_chunk_wrapper(chunk, if_not_compressed, recompress);
 
 	PG_RETURN_OID(uncompressed_chunk_id);
 }
@@ -1005,11 +843,7 @@ tsl_decompress_chunk(PG_FUNCTION_ARGS)
 				(errcode(ERRCODE_TS_HYPERTABLE_NOT_EXIST),
 				 errmsg("missing columnstore-enabled hypertable")));
 
-	if (ts_is_hypercore_am(uncompressed_chunk->amoid))
-	{
-		set_access_method(uncompressed_chunk_id, "heap");
-	}
-	else if (!ts_chunk_is_compressed(uncompressed_chunk))
+	if (!ts_chunk_is_compressed(uncompressed_chunk))
 	{
 		ereport((if_compressed ? NOTICE : ERROR),
 				(errcode(ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE),
