@@ -45,6 +45,7 @@
 #include "custom_type_cache.h"
 #include "guc.h"
 #include "hypertable_cache.h"
+#include "jsonb_utils.h"
 #include "trigger.h"
 #include "ts_catalog/array_utils.h"
 #include "ts_catalog/catalog.h"
@@ -75,8 +76,14 @@ is_sparse_index_type(const char *type)
 static void validate_hypertable_for_compression(Hypertable *ht);
 static List *build_columndefs(CompressionSettings *settings, Oid src_reloid);
 static ColumnDef *build_columndef_singlecolumn(const char *colname, Oid typid);
-static void compression_settings_update(Hypertable *ht, CompressionSettings *settings,
-										WithClauseResult *with_clause_options);
+static void compression_settings_set_manually_for_create(Hypertable *ht,
+														 CompressionSettings *settings,
+														 WithClauseResult *with_clause_options);
+static void compression_settings_set_manually_for_alter(Hypertable *ht,
+														CompressionSettings *settings,
+														WithClauseResult *with_clause_options);
+static void compression_settings_set_defaults(Hypertable *ht, CompressionSettings *settings,
+											  WithClauseResult *with_clause_options);
 
 static char *
 compression_column_segment_metadata_name(const char *type, int16 column_index)
@@ -170,7 +177,7 @@ compressed_column_metadata_attno(const CompressionSettings *settings, Oid chunk_
  * The heuristic for whether we should use the bloom filter sparse index.
  */
 static bool
-should_create_bloom_sparse_index(Form_pg_attribute attr, TypeCacheEntry *type, Oid src_reloid)
+should_create_bloom_sparse_index(Oid atttypid, TypeCacheEntry *type, Oid src_reloid)
 {
 	/*
 	 * The index must be enabled by the GUC.
@@ -185,7 +192,7 @@ should_create_bloom_sparse_index(Form_pg_attribute attr, TypeCacheEntry *type, O
 	 * which have better characteristics.
 	 */
 	FmgrInfo *finfo = NULL;
-	if (bloom1_get_hash_function(attr->atttypid, &finfo) == NULL)
+	if (bloom1_get_hash_function(atttypid, &finfo) == NULL)
 	{
 		return false;
 	}
@@ -197,8 +204,8 @@ should_create_bloom_sparse_index(Form_pg_attribute attr, TypeCacheEntry *type, O
 	 *    with the update time that is used as orderby.
 	 * This makes minmax indexes more suitable than bloom filters.
 	 */
-	if (attr->atttypid == TIMESTAMPTZOID || attr->atttypid == TIMESTAMPOID ||
-		attr->atttypid == TIMEOID || attr->atttypid == TIMETZOID || attr->atttypid == DATEOID)
+	if (atttypid == TIMESTAMPTZOID || atttypid == TIMESTAMPOID || atttypid == TIMEOID ||
+		atttypid == TIMETZOID || atttypid == DATEOID)
 	{
 		return false;
 	}
@@ -206,7 +213,7 @@ should_create_bloom_sparse_index(Form_pg_attribute attr, TypeCacheEntry *type, O
 	/*
 	 * For fractional arithmetic types, equality queries are unlikely.
 	 */
-	if (attr->atttypid == FLOAT4OID || attr->atttypid == FLOAT8OID || attr->atttypid == NUMERICOID)
+	if (atttypid == FLOAT4OID || atttypid == FLOAT8OID || atttypid == NUMERICOID)
 	{
 		return false;
 	}
@@ -221,19 +228,67 @@ should_create_bloom_sparse_index(Form_pg_attribute attr, TypeCacheEntry *type, O
 		return false;
 	}
 
-	/*
-	 * Bloom filter pushdown is not implemented for TAM at the moment, so keep
-	 * the old behavior with minmax sparse indexes. This check is actually not
-	 * enough, because the compressed chunk table is created when the
-	 * uncompressed chunk table is converted to TAM, and at this time it still
-	 * has the normal heap access method.
-	 */
-	if (ts_is_hypercore_am(ts_get_rel_am(src_reloid)))
+	return true;
+}
+
+static ColumnDef *
+create_sparse_index_column_def(Form_pg_attribute attr, const char *metadata_type)
+{
+	Assert(is_sparse_index_type(metadata_type));
+	Assert(strlen(metadata_type) <= 6);
+	ColumnDef *column_def = NULL;
+
+	const bool is_bloom = strcmp(metadata_type, "bloom1") == 0;
+
+	if (is_bloom)
 	{
-		return false;
+		/*
+		 * The type must be hashable. For some types we use our own hash functions
+		 * which have better characteristics.
+		 */
+		FmgrInfo *finfo = NULL;
+		if (bloom1_get_hash_function(attr->atttypid, &finfo) == NULL)
+			ereport(ERROR,
+					(errcode(ERRCODE_UNDEFINED_FUNCTION),
+					 errmsg("invalid bloom filter column type %s", format_type_be(attr->atttypid)),
+					 errdetail("Could not identify a hashing function for the type.")));
+		column_def =
+			makeColumnDef(compressed_column_metadata_name_v2(metadata_type, NameStr(attr->attname)),
+						  ts_custom_type_cache_get(CUSTOM_TYPE_BLOOM1)->type_oid,
+						  /* typmod = */ -1,
+						  /* collation = */ 0);
+
+		/*
+		 * We have our custom compression for bloom filters, and the
+		 * result is almost incompressible with lz4 (~2%), so disable it.
+		 */
+		column_def->storage = TYPSTORAGE_EXTERNAL;
+	}
+	else /* either min or max */
+	{
+		TypeCacheEntry *type = lookup_type_cache(attr->atttypid, TYPECACHE_LT_OPR);
+
+		/*
+		 * a comparison operator if required for min max operations
+		 */
+		if (!OidIsValid(type->lt_opr))
+			ereport(ERROR,
+					(errcode(ERRCODE_UNDEFINED_FUNCTION),
+					 errmsg("invalid minmax column type %s", format_type_be(attr->atttypid)),
+					 errdetail("Could not identify a less-than operator for the type.")));
+
+		column_def =
+			makeColumnDef(compressed_column_metadata_name_v2(metadata_type, NameStr(attr->attname)),
+						  attr->atttypid,
+						  attr->atttypmod,
+						  attr->attcollation);
+		if (attr->attstorage != TYPSTORAGE_PLAIN)
+		{
+			column_def->storage = TYPSTORAGE_MAIN;
+		}
 	}
 
-	return true;
+	return column_def;
 }
 
 /*
@@ -254,52 +309,6 @@ build_columndefs(CompressionSettings *settings, Oid src_reloid)
 
 	Relation rel = table_open(src_reloid, AccessShareLock);
 
-	Bitmapset *index_columns = NULL;
-	if (ts_guc_auto_sparse_indexes)
-	{
-		/*
-		 * Check which columns have btree indexes. We will create sparse minmax
-		 * indexes for them in compressed chunk.
-		 */
-		ListCell *lc;
-		List *index_oids = RelationGetIndexList(rel);
-		foreach (lc, index_oids)
-		{
-			Oid index_oid = lfirst_oid(lc);
-			Relation index_rel = index_open(index_oid, AccessShareLock);
-			IndexInfo *index_info = BuildIndexInfo(index_rel);
-			index_close(index_rel, NoLock);
-
-			/*
-			 * We want to create the sparse minmax index, if it can satisfy the same
-			 * kinds of queries as the uncompressed index. The simplest case is btree
-			 * which can satisfy equality and comparison tests, same as sparse minmax.
-			 *
-			 * If an uncompressed column has an index, we want to create a
-			 * sparse index for it as well. A sparse index can't satisfy ordering
-			 * queries, but at least we can use a bloom index to satisfy equality
-			 * queries. Create it when we have uncompressed index types that can
-			 * also satisfy equality.
-			 */
-			if (index_info->ii_Am != BTREE_AM_OID && index_info->ii_Am != HASH_AM_OID &&
-				index_info->ii_Am != BRIN_AM_OID)
-			{
-				continue;
-			}
-
-			for (int i = 0; i < index_info->ii_NumIndexKeyAttrs; i++)
-			{
-				const AttrNumber attno = index_info->ii_IndexAttrNumbers[i];
-				if (attno == InvalidAttrNumber)
-				{
-					continue;
-				}
-
-				index_columns = bms_add_member(index_columns, attno);
-			}
-		}
-	}
-
 	TupleDesc tupdesc = rel->rd_att;
 
 	for (int attoffset = 0; attoffset < tupdesc->natts; attoffset++)
@@ -362,68 +371,58 @@ build_columndefs(CompressionSettings *settings, Oid src_reloid)
 			def->storage = TYPSTORAGE_PLAIN;
 			compressed_column_defs = lappend(compressed_column_defs, def);
 		}
-		else if (bms_is_member(attr->attnum, index_columns))
+		else if (settings->fd.index)
 		{
-			TypeCacheEntry *type =
-				lookup_type_cache(attr->atttypid, TYPECACHE_LT_OPR | TYPECACHE_HASH_EXTENDED_PROC);
-
+			/* check sparse index columndefs is applicable */
+			bool is_bloom = ts_contains_sparse_index_config(settings,
+															NameStr(attr->attname),
+															ts_sparse_index_type_names
+																[_SparseIndexTypeEnumBloom]);
+			bool is_minmax = ts_contains_sparse_index_config(settings,
+															 NameStr(attr->attname),
+															 ts_sparse_index_type_names
+																 [_SparseIndexTypeEnumMinmax]);
 			/*
-			 * We can have various unusual user-defined types which do not
-			 * support comparison or hashing. The sparse indexes for the
-			 * non-orderby columns are not required for correctness, so just
-			 * don't create the sparse index if we lack the suitable operators.
+			 * We allow only one sparse index per column. Columns used in the ORDER BY
+			 * clause implicitly have a minmax index and adding a bloom filter on them is not
+			 * allowed.
+			 *
+			 * The parser is expected to enforce this constraint earlier, but we check again
+			 * here as a safeguard.
 			 */
-			bool can_use_minmax = OidIsValid(type->lt_opr);
+			Ensure((!is_bloom || !is_minmax),
+				   "Should not create bloom filter for minmax column \"%s\"",
+				   NameStr(attr->attname));
 
-			if (should_create_bloom_sparse_index(attr, type, src_reloid))
+			/* build sparse index columndefs if applicable */
+			if (is_bloom)
 			{
+				if (!ts_guc_enable_sparse_index_bloom)
+				{
+					ereport(ERROR,
+							(errcode(ERRCODE_OBJECT_NOT_IN_PREREQUISITE_STATE),
+							 errmsg("Creating bloom sparse index is disabled"),
+							 errhint("Set \"enable_sparse_index_bloom\" to true.")));
+				}
 				/*
 				 * Add bloom filter sparse index for this column.
 				 */
-				ColumnDef *bloom_column_def =
-					makeColumnDef(compressed_column_metadata_name_v2("bloom1",
-																	 NameStr(attr->attname)),
-								  ts_custom_type_cache_get(CUSTOM_TYPE_BLOOM1)->type_oid,
-								  /* typmod = */ -1,
-								  /* collation = */ 0);
-
-				/*
-				 * We have our custom compression for bloom filters, and the
-				 * result is almost incompressible with lz4 (~2%), so disable it.
-				 */
-				bloom_column_def->storage = TYPSTORAGE_EXTERNAL;
+				ColumnDef *bloom_column_def = create_sparse_index_column_def(attr, "bloom1");
 
 				compressed_column_defs = lappend(compressed_column_defs, bloom_column_def);
 			}
-			else if (can_use_minmax)
+			else if (is_minmax)
 			{
 				/*
 				 * Add minmax sparse index for this column.
 				 */
-				ColumnDef *def =
-					makeColumnDef(compressed_column_metadata_name_v2("min", NameStr(attr->attname)),
-								  attr->atttypid,
-								  attr->atttypmod,
-								  attr->attcollation);
-				if (attr->attstorage != TYPSTORAGE_PLAIN)
-				{
-					def->storage = TYPSTORAGE_MAIN;
-				}
+				ColumnDef *def = create_sparse_index_column_def(attr, "min");
 				compressed_column_defs = lappend(compressed_column_defs, def);
 
-				def =
-					makeColumnDef(compressed_column_metadata_name_v2("max", NameStr(attr->attname)),
-								  attr->atttypid,
-								  attr->atttypmod,
-								  attr->attcollation);
-				if (attr->attstorage != TYPSTORAGE_PLAIN)
-				{
-					def->storage = TYPSTORAGE_MAIN;
-				}
+				def = create_sparse_index_column_def(attr, "max");
 				compressed_column_defs = lappend(compressed_column_defs, def);
 			}
 		}
-
 		compressed_column_defs = lappend(compressed_column_defs,
 										 makeColumnDef(NameStr(attr->attname),
 													   compresseddata_oid,
@@ -545,6 +544,7 @@ create_compress_chunk(Hypertable *compress_ht, Chunk *src_chunk, Oid table_id)
 	 * is actually triggered allowing us to generate better compression
 	 * settings.
 	 */
+
 	if (!settings)
 	{
 		settings = ts_compression_settings_create(src_chunk->hypertable_relid,
@@ -552,19 +552,23 @@ create_compress_chunk(Hypertable *compress_ht, Chunk *src_chunk, Oid table_id)
 												  NULL,
 												  NULL,
 												  NULL,
+												  NULL,
 												  NULL);
-
-		Hypertable *ht = ts_hypertable_get_by_id(src_chunk->fd.hypertable_id);
-		compression_settings_update(ht, settings, ts_alter_table_with_clause_parse(NIL));
 	}
 
+	Hypertable *ht = ts_hypertable_get_by_id(src_chunk->fd.hypertable_id);
+	compression_settings_set_defaults(ht, settings, ts_alter_table_with_clause_parse(NIL));
+
 	if (OidIsValid(table_id))
 		compress_chunk->table_id = table_id;
 	else
 	{
 		List *column_defs = build_columndefs(settings, src_chunk->table_id);
-		compress_chunk->table_id =
-			compression_chunk_create(src_chunk, compress_chunk, column_defs, tablespace_oid);
+		compress_chunk->table_id = compression_chunk_create(src_chunk,
+															compress_chunk,
+															column_defs,
+															tablespace_oid,
+															settings);
 	}
 
 	if (!OidIsValid(compress_chunk->table_id))
@@ -695,7 +699,8 @@ validate_existing_constraints(Hypertable *ht, CompressionSettings *settings)
 				const char *attname = get_attname(settings->fd.relid, attnums[j], false);
 
 				/* is colno a segment-by or order_by column */
-				if (!form->conindid && !ts_array_is_member(settings->fd.segmentby, attname) &&
+				if (!form->conindid && (settings->fd.segmentby && settings->fd.orderby) &&
+					!ts_array_is_member(settings->fd.segmentby, attname) &&
 					!ts_array_is_member(settings->fd.orderby, attname))
 					ereport(WARNING,
 							(errmsg("column \"%s\" should be used for segmenting or ordering",
@@ -751,7 +756,8 @@ validate_existing_indexes(Hypertable *ht, CompressionSettings *settings)
 			if (attno == 0)
 				continue; /* skip check for expression column */
 			const char *attname = get_attname(ht->main_table_relid, attno, false);
-			if (!ts_array_is_member(settings->fd.segmentby, attname) &&
+			if ((settings->fd.segmentby && settings->fd.orderby) &&
+				!ts_array_is_member(settings->fd.segmentby, attname) &&
 				!ts_array_is_member(settings->fd.orderby, attname))
 				ereport(WARNING,
 						(errmsg("column \"%s\" should be used for segmenting or ordering",
@@ -901,6 +907,11 @@ tsl_process_compress_table(Hypertable *ht, WithClauseResult *with_clause_options
 		return disable_compression(ht, with_clause_options);
 	}
 
+	if (!with_clause_options[AlterTableFlagCompressChunkTimeInterval].is_default)
+	{
+		update_compress_chunk_time_interval(ht, with_clause_options);
+	}
+
 	settings = ts_compression_settings_get(ht->main_table_relid);
 	if (!settings)
 	{
@@ -909,10 +920,11 @@ tsl_process_compress_table(Hypertable *ht, WithClauseResult *with_clause_options
 												  NULL,
 												  NULL,
 												  NULL,
+												  NULL,
 												  NULL);
 	}
 
-	compression_settings_update(ht, settings, with_clause_options);
+	compression_settings_set_manually_for_alter(ht, settings, with_clause_options);
 
 	if (!TS_HYPERTABLE_HAS_COMPRESSION_TABLE(ht))
 	{
@@ -1035,20 +1047,16 @@ validate_hypertable_for_compression(Hypertable *ht)
 		Form_pg_trigger trigrec = (Form_pg_trigger) GETSTRUCT(tuple);
 
 		/*
-		 * We currently cannot support transition tables for DELETE triggers
-		 * on compressed tables that are not using hypercore table access
-		 * method since deleting a complete segment will not build a
+		 * We currently don't support transition tables for DELETE triggers
+		 * on compressed tables because deleting a complete segment will not build a
 		 * transition table for the delete.
 		 */
 		fastgetattr(tuple, Anum_pg_trigger_tgoldtable, pg_trigger->rd_att, &oldtable_isnull);
 		if (!oldtable_isnull && !TRIGGER_FOR_ROW(trigrec->tgtype) &&
-			TRIGGER_FOR_DELETE(trigrec->tgtype) && !ts_is_hypercore_am(ht->amoid))
+			TRIGGER_FOR_DELETE(trigrec->tgtype))
 			ereport(ERROR,
 					errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-					errmsg("DELETE triggers with transition tables not supported"),
-					errdetail(
-						"Compressed hypertables not using \"hypercore\" access method are not "
-						"supported if the trigger use transition tables."));
+					errmsg("DELETE triggers with transition tables not supported"));
 	}
 
 	systable_endscan(scan);
@@ -1120,7 +1128,7 @@ compression_setting_segmentby_get_default(const Hypertable *ht)
 	if (!isnull)
 	{
 		message = DatumGetTextPP(datum);
-		elog(WARNING,
+		elog(LOG_SERVER_ONLY,
 			 "there was some uncertainty picking the default segment by for the hypertable: %s",
 			 text_to_cstring(message));
 	}
@@ -1142,10 +1150,6 @@ compression_setting_segmentby_get_default(const Hypertable *ht)
 
 	initStringInfo(&result);
 	ts_array_append_stringinfo(column_res, &result);
-	elog(NOTICE,
-		 "default segment by for hypertable \"%s\" is set to \"%s\"",
-		 get_rel_name(ht->main_table_relid),
-		 result.data);
 
 	elog(LOG_SERVER_ONLY,
 		 "segment_by default: hypertable=\"%s\" columns=\"%s\" function: \"%s.%s\" confidence=%d",
@@ -1236,7 +1240,7 @@ compression_setting_orderby_get_default(Hypertable *ht, ArrayType *segmentby)
 	if (!isnull)
 	{
 		message = DatumGetTextPP(datum);
-		elog(WARNING,
+		elog(LOG_SERVER_ONLY,
 			 "there was some uncertainty picking the default order by for the hypertable: %s",
 			 text_to_cstring(message));
 	}
@@ -1260,17 +1264,6 @@ compression_setting_orderby_get_default(Hypertable *ht, ArrayType *segmentby)
 	else
 		orderby = "";
 
-	if (*orderby == '\0')
-		ereport(NOTICE,
-				(errmsg("default order by for hypertable \"%s\" is set to \"\"",
-						get_rel_name(ht->main_table_relid))),
-				errdetail("Segmentwise recompression will be disabled"));
-	else
-		elog(NOTICE,
-			 "default order by for hypertable \"%s\" is set to \"%s\"",
-			 get_rel_name(ht->main_table_relid),
-			 orderby);
-
 	elog(LOG_SERVER_ONLY,
 		 "order_by default: hypertable=\"%s\" clauses=\"%s\" function=\"%s.%s\" confidence=%d",
 		 get_rel_name(ht->main_table_relid),
@@ -1278,53 +1271,259 @@ compression_setting_orderby_get_default(Hypertable *ht, ArrayType *segmentby)
 		 get_namespace_name(get_func_namespace(orderby_fn)),
 		 get_func_name(orderby_fn),
 		 confidence);
+
+	if (*orderby == '\0')
+	{
+		return (OrderBySettings){ 0 };
+	}
+
 	return ts_compress_parse_order_collist(orderby, ht);
 }
 
+/* sparse indexes will only be set by default if there was no configuration */
+static bool
+can_set_default_sparse_index(CompressionSettings *settings)
+{
+	return (settings->fd.index == NULL) ||
+		   !ts_jsonb_has_key_value_str_field(settings->fd.index,
+											 ts_sparse_index_common_keys[SparseIndexKeySource],
+											 ts_sparse_index_source_names
+												 [_SparseIndexSourceEnumConfig]);
+}
+
+static Jsonb *
+compression_setting_sparse_index_get_default(Hypertable *ht, CompressionSettings *settings)
+{
+	bool has_object = false;
+	Bitmapset *sparse_index_columns = NULL;
+	JsonbParseState *parse_state = NULL;
+
+	/*
+	 * Sparse indexes are only created automatically if they are not set in compression settings
+	 */
+	if (!ts_guc_auto_sparse_indexes || !can_set_default_sparse_index(settings))
+		return NULL;
+
+	/*
+	 * Check which columns have btree indexes. We will create sparse minmax
+	 * indexes for them in compressed chunk.
+	 */
+	Relation rel = table_open(ht->main_table_relid, AccessShareLock);
+
+	ListCell *lc;
+	List *index_oids = RelationGetIndexList(rel);
+
+	pushJsonbValue(&parse_state, WJB_BEGIN_ARRAY, NULL);
+	foreach (lc, index_oids)
+	{
+		Oid index_oid = lfirst_oid(lc);
+		Relation index_rel = index_open(index_oid, AccessShareLock);
+		IndexInfo *index_info = BuildIndexInfo(index_rel);
+		index_close(index_rel, NoLock);
+
+		/*
+		 * We want to create the sparse minmax index, if it can satisfy the same
+		 * kinds of queries as the uncompressed index. The simplest case is btree
+		 * which can satisfy equality and comparison tests, same as sparse minmax.
+		 *
+		 * If an uncompressed column has an index, we want to create a
+		 * sparse index for it as well. A sparse index can't satisfy ordering
+		 * queries, but at least we can use a bloom index to satisfy equality
+		 * queries. Create it when we have uncompressed index types that can
+		 * also satisfy equality.
+		 */
+		if (index_info->ii_Am != BTREE_AM_OID && index_info->ii_Am != HASH_AM_OID &&
+			index_info->ii_Am != BRIN_AM_OID)
+		{
+			continue;
+		}
+
+		for (int i = 0; i < index_info->ii_NumIndexKeyAttrs; i++)
+		{
+			char *attname;
+			Oid atttypid;
+			SparseIndexConfig config;
+			TypeCacheEntry *type;
+			const AttrNumber attno = index_info->ii_IndexAttrNumbers[i];
+			if (attno == InvalidAttrNumber)
+			{
+				continue;
+			}
+			attname = get_attname(ht->main_table_relid, attno, false);
+			/* do not create sparse index for orderby columns */
+			if (ts_array_is_member(settings->fd.orderby, attname) ||
+				ts_array_is_member(settings->fd.segmentby, attname) ||
+				bms_is_member(attno, sparse_index_columns))
+				continue;
+
+			atttypid = get_atttype(ht->main_table_relid, attno);
+
+			type = lookup_type_cache(atttypid, TYPECACHE_LT_OPR | TYPECACHE_HASH_EXTENDED_PROC);
+
+			/* construct sparse index config */
+			if (ts_guc_enable_sparse_index_bloom &&
+				should_create_bloom_sparse_index(atttypid, type, ht->main_table_relid))
+				config.base.type = _SparseIndexTypeEnumBloom;
+			else if (OidIsValid(type->lt_opr))
+				config.base.type = _SparseIndexTypeEnumMinmax;
+			else
+				continue;
+
+			config.base.col = attname;
+			config.base.source = _SparseIndexSourceEnumDefault;
+
+			/* convert to json object */
+			ts_convert_sparse_index_config_to_jsonb(parse_state, &config);
+			sparse_index_columns = bms_add_member(sparse_index_columns, attno);
+			has_object = true;
+		}
+	}
+	table_close(rel, AccessShareLock);
+	return has_object ? JsonbValueToJsonb(pushJsonbValue(&parse_state, WJB_END_ARRAY, NULL)) : NULL;
+}
+
 static void
-compression_settings_update(Hypertable *ht, CompressionSettings *settings,
-							WithClauseResult *with_clause_options)
+compression_settings_set_defaults(Hypertable *ht, CompressionSettings *settings,
+								  WithClauseResult *with_clause_options)
 {
 	/* orderby arrays should always be in sync either all NULL or none */
 	Assert(
 		(settings->fd.orderby && settings->fd.orderby_desc && settings->fd.orderby_nullsfirst) ||
 		(!settings->fd.orderby && !settings->fd.orderby_desc && !settings->fd.orderby_nullsfirst));
 
-	if (!with_clause_options[AlterTableFlagCompressChunkTimeInterval].is_default)
+	bool add_orderby_sparse_index = false;
+	/* get default settings which will be stored at chunk level */
+	if (!(settings->fd.orderby) && with_clause_options[AlterTableFlagOrderBy].is_default)
 	{
-		update_compress_chunk_time_interval(ht, with_clause_options);
+		if (!settings->fd.segmentby && with_clause_options[AlterTableFlagSegmentBy].is_default)
+		{
+			settings->fd.segmentby = compression_setting_segmentby_get_default(ht);
+		}
+		settings->fd.index = ts_remove_orderby_sparse_index(settings);
+		OrderBySettings obs = compression_setting_orderby_get_default(ht, settings->fd.segmentby);
+		settings->fd.orderby = obs.orderby;
+		settings->fd.orderby_desc = obs.orderby_desc;
+		settings->fd.orderby_nullsfirst = obs.orderby_nullsfirst;
+		add_orderby_sparse_index = true;
 	}
 
+	if (ts_guc_auto_sparse_indexes && can_set_default_sparse_index(settings))
+	{
+		settings->fd.index = compression_setting_sparse_index_get_default(ht, settings);
+		settings->fd.index = ts_add_orderby_sparse_index(settings);
+	}
+	else if (add_orderby_sparse_index)
+	{
+		settings->fd.index = ts_add_orderby_sparse_index(settings);
+	}
+}
+
+static void
+compression_settings_set_manually_for_alter(Hypertable *ht, CompressionSettings *settings,
+											WithClauseResult *with_clause_options)
+{
+	/* orderby arrays should always be in sync either all NULL or none */
+	Assert(
+		(settings->fd.orderby && settings->fd.orderby_desc && settings->fd.orderby_nullsfirst) ||
+		(!settings->fd.orderby && !settings->fd.orderby_desc && !settings->fd.orderby_nullsfirst));
+
+	if (with_clause_options[AlterTableFlagSegmentBy].is_default &&
+		with_clause_options[AlterTableFlagOrderBy].is_default &&
+		with_clause_options[AlterTableFlagIndex].is_default)
+		return;
+
+	bool add_orderby_sparse_index = false;
 	if (!with_clause_options[AlterTableFlagSegmentBy].is_default)
 	{
 		settings->fd.segmentby =
 			ts_compress_hypertable_parse_segment_by(with_clause_options[AlterTableFlagSegmentBy],
 													ht);
 	}
-	else if (!settings->fd.segmentby && !settings->fd.orderby &&
-			 with_clause_options[AlterTableFlagOrderBy].is_default)
-	{
-		settings->fd.segmentby = compression_setting_segmentby_get_default(ht);
-	}
 
-	if (!with_clause_options[AlterTableFlagOrderBy].is_default || !settings->fd.orderby)
+	if (!with_clause_options[AlterTableFlagOrderBy].is_default)
 	{
-		OrderBySettings obs;
-		if (with_clause_options[AlterTableFlagOrderBy].is_default)
-		{
-			obs = compression_setting_orderby_get_default(ht, settings->fd.segmentby);
-		}
-		else
-		{
-			obs = ts_compress_hypertable_parse_order_by(with_clause_options[AlterTableFlagOrderBy],
-														ht);
-			obs = add_time_to_order_by_if_not_included(obs, settings->fd.segmentby, ht);
-		}
+		settings->fd.index = ts_remove_orderby_sparse_index(settings);
+		OrderBySettings obs =
+			ts_compress_hypertable_parse_order_by(with_clause_options[AlterTableFlagOrderBy], ht);
+		obs = add_time_to_order_by_if_not_included(obs, settings->fd.segmentby, ht);
 		settings->fd.orderby = obs.orderby;
 		settings->fd.orderby_desc = obs.orderby_desc;
 		settings->fd.orderby_nullsfirst = obs.orderby_nullsfirst;
+		add_orderby_sparse_index = true;
 	}
 
+	if (!with_clause_options[AlterTableFlagIndex].is_default)
+	{
+		if (!settings->fd.orderby)
+		{
+			ereport(ERROR,
+					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+					 errmsg("cannot set sparse index option without orderby option"),
+					 errdetail("Either set both options or remove both options to trigger default "
+							   "values")));
+		}
+		settings->fd.index =
+			ts_compress_hypertable_parse_index(with_clause_options[AlterTableFlagIndex], ht);
+		settings->fd.index = ts_add_orderby_sparse_index(settings);
+	}
+	else if (add_orderby_sparse_index)
+	{
+		settings->fd.index = ts_add_orderby_sparse_index(settings);
+	}
+
+	/* update manual settings */
+	ts_compression_settings_update(settings);
+}
+
+static void
+compression_settings_set_manually_for_create(Hypertable *ht, CompressionSettings *settings,
+											 WithClauseResult *with_clause_options)
+{
+	if (with_clause_options[CreateTableFlagSegmentBy].is_default &&
+		with_clause_options[CreateTableFlagOrderBy].is_default &&
+		with_clause_options[CreateTableFlagIndex].is_default)
+		return;
+
+	bool add_orderby_sparse_index = false;
+	if (!with_clause_options[CreateTableFlagSegmentBy].is_default)
+	{
+		settings->fd.segmentby =
+			ts_compress_hypertable_parse_segment_by(with_clause_options[CreateTableFlagSegmentBy],
+													ht);
+	}
+
+	if (!with_clause_options[CreateTableFlagOrderBy].is_default)
+	{
+		settings->fd.index = ts_remove_orderby_sparse_index(settings);
+		OrderBySettings obs =
+			ts_compress_hypertable_parse_order_by(with_clause_options[CreateTableFlagOrderBy], ht);
+		obs = add_time_to_order_by_if_not_included(obs, settings->fd.segmentby, ht);
+		settings->fd.orderby = obs.orderby;
+		settings->fd.orderby_desc = obs.orderby_desc;
+		settings->fd.orderby_nullsfirst = obs.orderby_nullsfirst;
+		add_orderby_sparse_index = true;
+	}
+
+	if (!with_clause_options[CreateTableFlagIndex].is_default)
+	{
+		if (!settings->fd.orderby)
+		{
+			ereport(ERROR,
+					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+					 errmsg("cannot set sparse index option without orderby option"),
+					 errdetail("Either set both options or remove both options to trigger default "
+							   "values")));
+		}
+		settings->fd.index =
+			ts_compress_hypertable_parse_index(with_clause_options[CreateTableFlagIndex], ht);
+		settings->fd.index = ts_add_orderby_sparse_index(settings);
+	}
+	else if (add_orderby_sparse_index)
+	{
+		settings->fd.index = ts_add_orderby_sparse_index(settings);
+	}
+
+	/* update manual settings */
 	ts_compression_settings_update(settings);
 }
 
@@ -1470,45 +1669,15 @@ tsl_compression_enable(Hypertable *ht, WithClauseResult *with_clause_options)
 	LockRelationOid(catalog_get_table_id(ts_catalog_get(), HYPERTABLE), RowExclusiveLock);
 	Oid ownerid = ts_rel_get_owner(ht->main_table_relid);
 	Oid tablespace_oid = get_rel_tablespace(ht->main_table_relid);
-	if (!with_clause_options[CreateTableFlagOrderBy].is_default ||
-		!with_clause_options[CreateTableFlagSegmentBy].is_default)
-	{
-		CompressionSettings *settings = ts_compression_settings_create(ht->main_table_relid,
-																	   InvalidOid,
-																	   NULL,
-																	   NULL,
-																	   NULL,
-																	   NULL);
+	CompressionSettings *settings = ts_compression_settings_create(ht->main_table_relid,
+																   InvalidOid,
+																   NULL,
+																   NULL,
+																   NULL,
+																   NULL,
+																   NULL);
 
-		if (!with_clause_options[CreateTableFlagSegmentBy].is_default)
-		{
-			settings->fd.segmentby =
-				ts_compress_hypertable_parse_segment_by(with_clause_options
-															[CreateTableFlagSegmentBy],
-														ht);
-		}
-
-		if (!with_clause_options[CreateTableFlagOrderBy].is_default || !settings->fd.orderby)
-		{
-			OrderBySettings obs;
-			if (with_clause_options[CreateTableFlagOrderBy].is_default)
-			{
-				obs = compression_setting_orderby_get_default(ht, settings->fd.segmentby);
-			}
-			else
-			{
-				obs = ts_compress_hypertable_parse_order_by(with_clause_options
-																[CreateTableFlagOrderBy],
-															ht);
-				obs = add_time_to_order_by_if_not_included(obs, settings->fd.segmentby, ht);
-			}
-			settings->fd.orderby = obs.orderby;
-			settings->fd.orderby_desc = obs.orderby_desc;
-			settings->fd.orderby_nullsfirst = obs.orderby_nullsfirst;
-		}
-
-		ts_compression_settings_update(settings);
-	}
+	compression_settings_set_manually_for_create(ht, settings, with_clause_options);
 	int compress_htid = compression_hypertable_create(ht, ownerid, tablespace_oid);
 	ts_hypertable_set_compressed(ht, compress_htid);
 }
