@@ -41,5 +41,4 @@ void continuous_agg_update_materialization(Hypertable *mat_ht, const ContinuousA
 										   SchemaAndName partial_view,
 										   SchemaAndName materialization_table,
 										   const NameData *time_column_name,
-										   InternalTimeRange new_materialization_range,
-										   InternalTimeRange invalidation_range, int32 chunk_id);
+										   InternalTimeRange materialization_range, int32 chunk_id);
