@@ -5,8 +5,6 @@
  */
 #include <postgres.h>
 
-#include "bgw/job.h"
-#include "bgw_policy/policies_v2.h"
 #include <access/xact.h>
 #include <executor/spi.h>
 #include <fmgr.h>
@@ -20,6 +18,9 @@
 #include <utils/lsyscache.h>
 #include <utils/snapmgr.h>
 
+#include "bgw_policy/policies_v2.h"
+#include "continuous_aggs/invalidation_multi.h"
+#include "debug_point.h"
 #include "dimension.h"
 #include "dimension_slice.h"
 #include "guc.h"
@@ -40,13 +41,13 @@
 		 LOG :                                                                                     \
 		 DEBUG1)
 
-typedef struct CaggRefreshState
+typedef struct ContinuousAggRefreshState
 {
 	ContinuousAgg cagg;
 	Hypertable *cagg_ht;
 	InternalTimeRange refresh_window;
 	SchemaAndName partial_view;
-} CaggRefreshState;
+} ContinuousAggRefreshState;
 
 static Hypertable *cagg_get_hypertable_or_fail(int32 hypertable_id);
 static InternalTimeRange get_largest_bucketed_window(Oid timetype, int64 bucket_width);
@@ -54,40 +55,63 @@ static InternalTimeRange
 compute_inscribed_bucketed_refresh_window(const ContinuousAgg *cagg,
 										  const InternalTimeRange *const refresh_window,
 										  const int64 bucket_width);
-static InternalTimeRange
-compute_circumscribed_bucketed_refresh_window(const ContinuousAgg *cagg,
-											  const InternalTimeRange *const refresh_window,
-											  const ContinuousAggsBucketFunction *bucket_function);
-static void continuous_agg_refresh_init(CaggRefreshState *refresh, const ContinuousAgg *cagg,
+static void continuous_agg_refresh_init(ContinuousAggRefreshState *refresh,
+										const ContinuousAgg *cagg,
 										const InternalTimeRange *refresh_window);
-static void continuous_agg_refresh_execute(const CaggRefreshState *refresh,
+static void continuous_agg_refresh_execute(const ContinuousAggRefreshState *refresh,
 										   const InternalTimeRange *bucketed_refresh_window,
 										   const int32 chunk_id);
 static void log_refresh_window(int elevel, const ContinuousAgg *cagg,
-							   const InternalTimeRange *refresh_window, const char *msg,
-							   CaggRefreshContext context);
+							   const InternalTimeRange *refresh_window,
+							   ContinuousAggRefreshContext context);
 static void continuous_agg_refresh_execute_wrapper(const InternalTimeRange *bucketed_refresh_window,
-												   const CaggRefreshContext context,
+												   const ContinuousAggRefreshContext context,
 												   const long iteration, void *arg1_refresh,
 												   void *arg2_chunk_id);
-static void update_merged_refresh_window(const InternalTimeRange *bucketed_refresh_window,
-										 const CaggRefreshContext context, const long iteration,
-										 void *arg1_merged_refresh_window, void *arg2);
 static void continuous_agg_refresh_with_window(const ContinuousAgg *cagg,
 											   const InternalTimeRange *refresh_window,
 											   const InvalidationStore *invalidations,
-											   int32 chunk_id, const bool do_merged_refresh,
-											   const InternalTimeRange merged_refresh_window,
-											   const CaggRefreshContext context);
-static void emit_up_to_date_notice(const ContinuousAgg *cagg, const CaggRefreshContext context);
+											   int32 chunk_id,
+											   const ContinuousAggRefreshContext context);
+static void emit_up_to_date_notice(const ContinuousAgg *cagg,
+								   const ContinuousAggRefreshContext context);
 static bool process_cagg_invalidations_and_refresh(const ContinuousAgg *cagg,
 												   const InternalTimeRange *refresh_window,
-												   const CaggRefreshContext context, int32 chunk_id,
-												   bool force);
+												   const ContinuousAggRefreshContext context,
+												   int32 chunk_id, bool force);
 static void fill_bucket_offset_origin(const ContinuousAgg *cagg,
 									  const InternalTimeRange *const refresh_window,
 									  NullableDatum *offset, NullableDatum *origin);
 
+/*
+ * Get all hypertables that are using WAL-invalidations.
+ */
+static List *
+get_all_wal_using_hypertables(void)
+{
+	ScanIterator iterator =
+		ts_scan_iterator_create(CONTINUOUS_AGG, AccessShareLock, CurrentMemoryContext);
+	List *hypertables = NIL;
+
+	/* Collect OID of all tables using continuous aggregates */
+	ts_scanner_foreach(&iterator)
+	{
+		bool isnull;
+		Datum datum = slot_getattr(ts_scan_iterator_slot(&iterator),
+								   Anum_continuous_agg_raw_hypertable_id,
+								   &isnull);
+
+		Assert(!isnull);
+		int32 hypertable_id = DatumGetInt32(datum);
+		Oid relid = ts_hypertable_id_to_relid(hypertable_id, false);
+		if (!has_invalidation_trigger(relid))
+			hypertables = list_append_unique_int(hypertables, hypertable_id);
+	}
+	ts_scan_iterator_close(&iterator);
+
+	return hypertables;
+}
+
 static Hypertable *
 cagg_get_hypertable_or_fail(int32 hypertable_id)
 {
@@ -209,7 +233,7 @@ compute_inscribed_bucketed_refresh_window(const ContinuousAgg *cagg,
  * Get the offset as Datum value of an integer based bucket
  */
 static Datum
-int_bucket_offset_to_datum(Oid type, const ContinuousAggsBucketFunction *bucket_function)
+int_bucket_offset_to_datum(Oid type, const ContinuousAggBucketFunction *bucket_function)
 {
 	Assert(bucket_function->bucket_time_based == false);
 
@@ -302,10 +326,10 @@ fill_bucket_offset_origin(const ContinuousAgg *cagg, const InternalTimeRange *co
  * The circumscribed behaviour is also used for a refresh on drop, when the refresh is called during
  * dropping chunks manually or as part of retention policy.
  */
-static InternalTimeRange
+InternalTimeRange
 compute_circumscribed_bucketed_refresh_window(const ContinuousAgg *cagg,
 											  const InternalTimeRange *const refresh_window,
-											  const ContinuousAggsBucketFunction *bucket_function)
+											  const ContinuousAggBucketFunction *bucket_function)
 {
 	Assert(cagg != NULL);
 	Assert(cagg->bucket_function != NULL);
@@ -384,7 +408,7 @@ compute_circumscribed_bucketed_refresh_window(const ContinuousAgg *cagg,
  * The state holds information for executing a refresh of a continuous aggregate.
  */
 static void
-continuous_agg_refresh_init(CaggRefreshState *refresh, const ContinuousAgg *cagg,
+continuous_agg_refresh_init(ContinuousAggRefreshState *refresh, const ContinuousAgg *cagg,
 							const InternalTimeRange *refresh_window)
 {
 	MemSet(refresh, 0, sizeof(*refresh));
@@ -402,7 +426,7 @@ continuous_agg_refresh_init(CaggRefreshState *refresh, const ContinuousAgg *cagg
  * refresh state.
  */
 static void
-continuous_agg_refresh_execute(const CaggRefreshState *refresh,
+continuous_agg_refresh_execute(const ContinuousAggRefreshState *refresh,
 							   const InternalTimeRange *bucketed_refresh_window,
 							   const int32 chunk_id)
 {
@@ -410,14 +434,6 @@ continuous_agg_refresh_execute(const CaggRefreshState *refresh,
 		.schema = &refresh->cagg_ht->fd.schema_name,
 		.name = &refresh->cagg_ht->fd.table_name,
 	};
-	/* The materialization function takes two ranges, one for new data and one
-	 * for invalidated data. A refresh just uses one of them so the other one
-	 * has a zero range. */
-	InternalTimeRange unused_invalidation_range = {
-		.type = refresh->refresh_window.type,
-		.start = 0,
-		.end = 0,
-	};
 	const Dimension *time_dim = hyperspace_get_open_dimension(refresh->cagg_ht->space, 0);
 
 	Assert(time_dim != NULL);
@@ -428,31 +444,21 @@ continuous_agg_refresh_execute(const CaggRefreshState *refresh,
 										  cagg_hypertable_name,
 										  &time_dim->fd.column_name,
 										  *bucketed_refresh_window,
-										  unused_invalidation_range,
 										  chunk_id);
 }
 
 static void
 log_refresh_window(int elevel, const ContinuousAgg *cagg, const InternalTimeRange *refresh_window,
-				   const char *msg, CaggRefreshContext context)
+				   ContinuousAggRefreshContext context)
 {
-	Datum start_ts;
-	Datum end_ts;
-	Oid outfuncid = InvalidOid;
-	bool isvarlena;
-
-	start_ts = ts_internal_to_time_value(refresh_window->start, refresh_window->type);
-	end_ts = ts_internal_to_time_value(refresh_window->end, refresh_window->type);
-	getTypeOutputInfo(refresh_window->type, &outfuncid, &isvarlena);
-	Assert(!isvarlena);
-
+	const char *msg = "continuous aggregate refresh (individual invalidation) on";
 	if (context.callctx == CAGG_REFRESH_POLICY_BATCHED)
 		elog(elevel,
 			 "%s \"%s\" in window [ %s, %s ] (batch %d of %d)",
 			 msg,
 			 NameStr(cagg->data.user_view_name),
-			 DatumGetCString(OidFunctionCall1(outfuncid, start_ts)),
-			 DatumGetCString(OidFunctionCall1(outfuncid, end_ts)),
+			 ts_internal_to_time_string(refresh_window->start, refresh_window->type),
+			 ts_internal_to_time_string(refresh_window->end, refresh_window->type),
 			 context.processing_batch,
 			 context.number_of_batches);
 	else
@@ -460,58 +466,34 @@ log_refresh_window(int elevel, const ContinuousAgg *cagg, const InternalTimeRang
 			 "%s \"%s\" in window [ %s, %s ]",
 			 msg,
 			 NameStr(cagg->data.user_view_name),
-			 DatumGetCString(OidFunctionCall1(outfuncid, start_ts)),
-			 DatumGetCString(OidFunctionCall1(outfuncid, end_ts)));
+			 ts_internal_to_time_string(refresh_window->start, refresh_window->type),
+			 ts_internal_to_time_string(refresh_window->end, refresh_window->type));
 }
 
 typedef void (*scan_refresh_ranges_funct_t)(const InternalTimeRange *bucketed_refresh_window,
-											const CaggRefreshContext context,
+											const ContinuousAggRefreshContext context,
 											const long iteration, /* 0 is first range */
 											void *arg1, void *arg2);
 
 static void
 continuous_agg_refresh_execute_wrapper(const InternalTimeRange *bucketed_refresh_window,
-									   const CaggRefreshContext context, const long iteration,
-									   void *arg1_refresh, void *arg2_chunk_id)
+									   const ContinuousAggRefreshContext context,
+									   const long iteration, void *arg1_refresh,
+									   void *arg2_chunk_id)
 {
-	const CaggRefreshState *refresh = (const CaggRefreshState *) arg1_refresh;
+	const ContinuousAggRefreshState *refresh = (const ContinuousAggRefreshState *) arg1_refresh;
 	const int32 chunk_id = *(const int32 *) arg2_chunk_id;
 	(void) iteration;
 
-	log_refresh_window(CAGG_REFRESH_LOG_LEVEL,
-					   &refresh->cagg,
-					   bucketed_refresh_window,
-					   "continuous aggregate refresh (individual invalidation) on",
-					   context);
+	log_refresh_window(CAGG_REFRESH_LOG_LEVEL, &refresh->cagg, bucketed_refresh_window, context);
 	continuous_agg_refresh_execute(refresh, bucketed_refresh_window, chunk_id);
 }
 
-static void
-update_merged_refresh_window(const InternalTimeRange *bucketed_refresh_window,
-							 const CaggRefreshContext context, const long iteration,
-							 void *arg1_merged_refresh_window, void *arg2)
-{
-	InternalTimeRange *merged_refresh_window = (InternalTimeRange *) arg1_merged_refresh_window;
-	(void) arg2;
-
-	if (iteration == 0)
-		*merged_refresh_window = *bucketed_refresh_window;
-	else
-	{
-		if (bucketed_refresh_window->start < merged_refresh_window->start)
-			merged_refresh_window->start = bucketed_refresh_window->start;
-
-		if (bucketed_refresh_window->end > merged_refresh_window->end)
-			merged_refresh_window->end = bucketed_refresh_window->end;
-	}
-}
-
 static long
 continuous_agg_scan_refresh_window_ranges(const ContinuousAgg *cagg,
 										  const InternalTimeRange *refresh_window,
 										  const InvalidationStore *invalidations,
-										  const ContinuousAggsBucketFunction *bucket_function,
-										  const CaggRefreshContext context,
+										  const ContinuousAggRefreshContext context,
 										  scan_refresh_ranges_funct_t exec_func, void *func_arg1,
 										  void *func_arg2)
 {
@@ -544,7 +526,9 @@ continuous_agg_scan_refresh_window_ranges(const ContinuousAgg *cagg,
 		};
 
 		InternalTimeRange bucketed_refresh_window =
-			compute_circumscribed_bucketed_refresh_window(cagg, &invalidation, bucket_function);
+			compute_circumscribed_bucketed_refresh_window(cagg,
+														  &invalidation,
+														  cagg->bucket_function);
 
 		(*exec_func)(&bucketed_refresh_window, context, count, func_arg1, func_arg2);
 
@@ -586,11 +570,9 @@ static void
 continuous_agg_refresh_with_window(const ContinuousAgg *cagg,
 								   const InternalTimeRange *refresh_window,
 								   const InvalidationStore *invalidations, int32 chunk_id,
-								   const bool do_merged_refresh,
-								   const InternalTimeRange merged_refresh_window,
-								   const CaggRefreshContext context)
+								   const ContinuousAggRefreshContext context)
 {
-	CaggRefreshState refresh;
+	ContinuousAggRefreshState refresh;
 
 	continuous_agg_refresh_init(&refresh, cagg, refresh_window);
 
@@ -607,35 +589,15 @@ continuous_agg_refresh_with_window(const ContinuousAgg *cagg,
 	if (ContinuousAggIsFinalized(cagg))
 		chunk_id = INVALID_CHUNK_ID;
 
-	if (do_merged_refresh)
-	{
-		Assert(merged_refresh_window.type == refresh_window->type);
-		Assert(merged_refresh_window.start >= refresh_window->start);
-		Assert((cagg->bucket_function->bucket_fixed_interval == false) ||
-			   (merged_refresh_window.end -
-					ts_continuous_agg_fixed_bucket_width(cagg->bucket_function) <=
-				refresh_window->end));
-
-		log_refresh_window(CAGG_REFRESH_LOG_LEVEL,
-						   cagg,
-						   &merged_refresh_window,
-						   "continuous aggregate refresh (merged invalidation) on",
-						   context);
-		continuous_agg_refresh_execute(&refresh, &merged_refresh_window, chunk_id);
-	}
-	else
-	{
-		long count pg_attribute_unused();
-		count = continuous_agg_scan_refresh_window_ranges(cagg,
-														  refresh_window,
-														  invalidations,
-														  cagg->bucket_function,
-														  context,
-														  continuous_agg_refresh_execute_wrapper,
-														  (void *) &refresh /* arg1 */,
-														  (void *) &chunk_id /* arg2 */);
-		Assert(count);
-	}
+	long count pg_attribute_unused();
+	count = continuous_agg_scan_refresh_window_ranges(cagg,
+													  refresh_window,
+													  invalidations,
+													  context,
+													  continuous_agg_refresh_execute_wrapper,
+													  (void *) &refresh /* arg1 */,
+													  (void *) &chunk_id /* arg2 */);
+	Assert(count);
 }
 
 #define REFRESH_FUNCTION_NAME "refresh_continuous_aggregate()"
@@ -685,7 +647,7 @@ continuous_agg_refresh(PG_FUNCTION_ARGS)
 	else
 		refresh_window.end = ts_time_get_noend_or_max(refresh_window.type);
 
-	CaggRefreshContext context = { .callctx = CAGG_REFRESH_WINDOW };
+	ContinuousAggRefreshContext context = { .callctx = CAGG_REFRESH_WINDOW };
 	continuous_agg_refresh_internal(cagg,
 									&refresh_window,
 									context,
@@ -700,7 +662,7 @@ continuous_agg_refresh(PG_FUNCTION_ARGS)
 }
 
 static void
-emit_up_to_date_notice(const ContinuousAgg *cagg, const CaggRefreshContext context)
+emit_up_to_date_notice(const ContinuousAgg *cagg, const ContinuousAggRefreshContext context)
 {
 	switch (context.callctx)
 	{
@@ -716,54 +678,36 @@ emit_up_to_date_notice(const ContinuousAgg *cagg, const CaggRefreshContext conte
 	}
 }
 
-void
-continuous_agg_calculate_merged_refresh_window(const ContinuousAgg *cagg,
-											   const InternalTimeRange *refresh_window,
-											   const InvalidationStore *invalidations,
-											   const ContinuousAggsBucketFunction *bucket_function,
-											   InternalTimeRange *merged_refresh_window,
-											   const CaggRefreshContext context)
-{
-	long count pg_attribute_unused();
-	count = continuous_agg_scan_refresh_window_ranges(cagg,
-													  refresh_window,
-													  invalidations,
-													  bucket_function,
-													  context,
-													  update_merged_refresh_window,
-													  (void *) merged_refresh_window,
-													  NULL /* arg2 */);
-	Assert(count);
-}
-
 static bool
 process_cagg_invalidations_and_refresh(const ContinuousAgg *cagg,
 									   const InternalTimeRange *refresh_window,
-									   const CaggRefreshContext context, int32 chunk_id, bool force)
+									   const ContinuousAggRefreshContext context, int32 chunk_id,
+									   bool force)
 {
 	InvalidationStore *invalidations;
 	Oid hyper_relid = ts_hypertable_id_to_relid(cagg->data.mat_hypertable_id, false);
-	bool do_merged_refresh = false;
-	InternalTimeRange merged_refresh_window;
 
-	/* Lock the continuous aggregate's materialized hypertable to protect
-	 * against concurrent refreshes. Only concurrent reads will be
-	 * allowed. This is a heavy lock that serializes all refreshes on the same
-	 * continuous aggregate. We might want to consider relaxing this in the
-	 * future, e.g., we'd like to at least allow concurrent refreshes on the
-	 * same continuous aggregate when they don't have overlapping refresh
-	 * windows.
+	/* Lock the continuous aggregate's materialized hypertable to protect against
+	 * concurrent invalidation log processing.
+	 *
+	 * It will produce rows in the `continuous_aggs_materialization_ranges` table
+	 * to be materialized later either serially or in parallel for non-overlap
+	 * refresh ranges.
+	 *
+	 * This is supposed to be a short transaction and in the future we can consider
+	 * relaxing this lock.
 	 */
-	LockRelationOid(hyper_relid, RowExclusiveLock);
+	LockRelationOid(hyper_relid, ExclusiveLock);
 	invalidations = invalidation_process_cagg_log(cagg,
 												  refresh_window,
 												  ts_guc_cagg_max_individual_materializations,
-												  &do_merged_refresh,
-												  &merged_refresh_window,
 												  context,
 												  force);
+	SPI_commit_and_chain();
 
-	if (invalidations != NULL || do_merged_refresh)
+	DEBUG_WAITPOINT("after_process_cagg_invalidations_for_refresh_lock");
+
+	if (invalidations != NULL)
 	{
 		if (context.callctx == CAGG_REFRESH_CREATION)
 		{
@@ -774,13 +718,7 @@ process_cagg_invalidations_and_refresh(const ContinuousAgg *cagg,
 							 "aggregate on creation.")));
 		}
 
-		continuous_agg_refresh_with_window(cagg,
-										   refresh_window,
-										   invalidations,
-										   chunk_id,
-										   do_merged_refresh,
-										   merged_refresh_window,
-										   context);
+		continuous_agg_refresh_with_window(cagg, refresh_window, invalidations, chunk_id, context);
 		if (invalidations)
 			invalidation_store_free(invalidations);
 		return true;
@@ -792,7 +730,7 @@ process_cagg_invalidations_and_refresh(const ContinuousAgg *cagg,
 void
 continuous_agg_refresh_internal(const ContinuousAgg *cagg,
 								const InternalTimeRange *refresh_window_arg,
-								const CaggRefreshContext context, const bool start_isnull,
+								const ContinuousAggRefreshContext context, const bool start_isnull,
 								const bool end_isnull, bool bucketing_refresh_window, bool force,
 								bool process_hypertable_invalidations, bool extend_last_bucket)
 {
@@ -949,7 +887,32 @@ continuous_agg_refresh_internal(const ContinuousAgg *cagg,
 	}
 
 	if (process_hypertable_invalidations)
-		invalidation_process_hypertable_log(cagg->data.raw_hypertable_id, refresh_window.type);
+	{
+		/*
+		 * If we are using trigger-based invalidations, we can process the
+		 * invalidations for the associated hypertable only and later read the
+		 * invalidations for other hypertables, but when using WAL-based
+		 * invalidation we need to process all of the hypertables that are
+		 * currently using WAL.
+		 *
+		 * We want to prevent any changes to how invalidations are collected
+		 * in the meantime since changing the invalidation collection method
+		 * while this is running might cause problems and miss invalidations.
+		 *
+		 * Concurrency on the replication slot is controlled using some
+		 * special sauce in ReplicationSlotAcquire(), which is called inside
+		 * pg_logical_slot_get_changes_guts().
+		 *
+		 * This will currently generate an error rather than blocking on the
+		 * lock, so we need to add a separate lock to ensure a blocking
+		 * behaviour.
+		 */
+		if (has_invalidation_trigger(
+				ts_hypertable_id_to_relid(cagg->data.raw_hypertable_id, false)))
+			invalidation_process_hypertable_log(cagg->data.raw_hypertable_id, refresh_window.type);
+		else
+			multi_invalidation_process_hypertable_log(get_all_wal_using_hypertables());
+	}
 
 	/* Commit and Start a new transaction */
 	SPI_commit_and_chain();
@@ -980,27 +943,17 @@ static void
 debug_refresh_window(const ContinuousAgg *cagg, const InternalTimeRange *refresh_window,
 					 const char *msg)
 {
-	Datum start_ts;
-	Datum end_ts;
-	Oid outfuncid = InvalidOid;
-	bool isvarlena;
-
-	start_ts = ts_internal_to_time_value(refresh_window->start, refresh_window->type);
-	end_ts = ts_internal_to_time_value(refresh_window->end, refresh_window->type);
-	getTypeOutputInfo(refresh_window->type, &outfuncid, &isvarlena);
-	Assert(!isvarlena);
-
 	elog(DEBUG1,
 		 "%s \"%s\" in window [ %s, %s ] internal [ " INT64_FORMAT ", " INT64_FORMAT
 		 " ] minimum [ %s ]",
 		 msg,
 		 NameStr(cagg->data.user_view_name),
-		 DatumGetCString(OidFunctionCall1(outfuncid, start_ts)),
-		 DatumGetCString(OidFunctionCall1(outfuncid, end_ts)),
+		 ts_internal_to_time_string(refresh_window->start, refresh_window->type),
+		 ts_internal_to_time_string(refresh_window->end, refresh_window->type),
 		 refresh_window->start,
 		 refresh_window->end,
-		 DatumGetCString(
-			 OidFunctionCall1(outfuncid, Int64GetDatum(ts_time_get_min(refresh_window->type)))));
+		 ts_datum_to_string(Int64GetDatum(ts_time_get_min(refresh_window->type)),
+							refresh_window->type));
 }
 
 List *
@@ -1044,8 +997,7 @@ continuous_agg_split_refresh_window(ContinuousAgg *cagg, InternalTimeRange *orig
 		{
 			elog(LOG,
 				 "no min slice range start for continuous aggregate \"%s.%s\", falling back to "
-				 "single "
-				 "batch processing",
+				 "single batch processing",
 				 NameStr(cagg->data.user_view_schema),
 				 NameStr(cagg->data.user_view_name));
 			return NIL;
@@ -1097,18 +1049,12 @@ continuous_agg_split_refresh_window(ContinuousAgg *cagg, InternalTimeRange *orig
 		Oid type = IS_TIMESTAMP_TYPE(refresh_window.type) ? INTERVALOID : refresh_window.type;
 		Datum refresh_size_interval = ts_internal_to_interval_value(refresh_window_size, type);
 		Datum batch_size_interval = ts_internal_to_interval_value(batch_size, type);
-		Oid typoutputfunc;
-		bool isvarlena;
-		FmgrInfo typoutputinfo;
-
-		getTypeOutputInfo(type, &typoutputfunc, &isvarlena);
-		fmgr_info(typoutputfunc, &typoutputinfo);
 
 		elog(LOG,
 			 "refresh window size (%s) is smaller than or equal to batch size (%s), falling back "
 			 "to single batch processing",
-			 OutputFunctionCall(&typoutputinfo, refresh_size_interval),
-			 OutputFunctionCall(&typoutputinfo, batch_size_interval));
+			 ts_datum_to_string(refresh_size_interval, type),
+			 ts_datum_to_string(batch_size_interval, type));
 		return NIL;
 	}
 
