@@ -27,12 +27,28 @@ typedef struct QualPushdownContext
 	RelOptInfo *compressed_rel;
 	RangeTblEntry *chunk_rte;
 	RangeTblEntry *compressed_rte;
+	CompressionSettings *settings;
+
+	/*
+	 * This is actually the result, not the static input context like above, but
+	 * there's no way to separate this properly using the expression tree mutator
+	 * interface.
+	 */
 	bool can_pushdown;
 	bool needs_recheck;
-	CompressionSettings *settings;
 } QualPushdownContext;
 
-static Node *modify_expression(Node *node, QualPushdownContext *context);
+static QualPushdownContext
+copy_context(const QualPushdownContext *source)
+{
+	QualPushdownContext copy;
+	copy = *source;
+	copy.can_pushdown = true;
+	copy.needs_recheck = false;
+	return copy;
+}
+
+static Node *qual_pushdown_mutator(Node *node, QualPushdownContext *context);
 
 void
 pushdown_quals(PlannerInfo *root, CompressionSettings *settings, RelOptInfo *chunk_rel,
@@ -40,7 +56,7 @@ pushdown_quals(PlannerInfo *root, CompressionSettings *settings, RelOptInfo *chu
 {
 	ListCell *lc;
 	List *decompress_clauses = NIL;
-	QualPushdownContext context = {
+	QualPushdownContext base_context = {
 		.chunk_rel = chunk_rel,
 		.compressed_rel = compressed_rel,
 		.chunk_rte = planner_rt_fetch(chunk_rel->relid, root),
@@ -51,20 +67,11 @@ pushdown_quals(PlannerInfo *root, CompressionSettings *settings, RelOptInfo *chu
 	foreach (lc, chunk_rel->baserestrictinfo)
 	{
 		RestrictInfo *ri = lfirst(lc);
-		Expr *expr;
 
-		/* pushdown is not safe for volatile expressions */
-		if (contain_volatile_functions((Node *) ri->clause))
-		{
-			decompress_clauses = lappend(decompress_clauses, ri);
-			continue;
-		}
+		QualPushdownContext clause_context = copy_context(&base_context);
+		Node *pushed_down = qual_pushdown_mutator((Node *) ri->clause, &clause_context);
 
-		context.can_pushdown = true;
-		context.needs_recheck = false;
-		expr = (Expr *) modify_expression((Node *) ri->clause, &context);
-
-		if (context.can_pushdown)
+		if (clause_context.can_pushdown)
 		{
 			/*
 			 * We have to call eval_const_expressions after pushing down
@@ -73,13 +80,13 @@ pushdown_quals(PlannerInfo *root, CompressionSettings *settings, RelOptInfo *chu
 			 * allowed to have nested AND boolexprs. They break some functions like
 			 * generate_bitmap_or_paths().
 			 */
-			expr = (Expr *) eval_const_expressions(root, (Node *) expr);
+			pushed_down = eval_const_expressions(root, pushed_down);
 
-			if (IsA(expr, BoolExpr) && ((BoolExpr *) expr)->boolop == AND_EXPR)
+			if (IsA(pushed_down, BoolExpr) && castNode(BoolExpr, pushed_down)->boolop == AND_EXPR)
 			{
 				/* have to separate out and expr into different restrict infos */
 				ListCell *lc_and;
-				BoolExpr *bool_expr = (BoolExpr *) expr;
+				BoolExpr *bool_expr = castNode(BoolExpr, pushed_down);
 				foreach (lc_and, bool_expr->args)
 				{
 					compressed_rel->baserestrictinfo =
@@ -89,11 +96,15 @@ pushdown_quals(PlannerInfo *root, CompressionSettings *settings, RelOptInfo *chu
 			}
 			else
 				compressed_rel->baserestrictinfo =
-					lappend(compressed_rel->baserestrictinfo, make_simple_restrictinfo(root, expr));
+					lappend(compressed_rel->baserestrictinfo,
+							make_simple_restrictinfo(root, (Expr *) pushed_down));
 		}
-		/* We need to check the restriction clause on the decompress node if the clause can't be
-		 * pushed down or needs re-checking */
-		if (!context.can_pushdown || context.needs_recheck || chunk_partial)
+
+		/*
+		 * We need to check the restriction clause on the decompress node if the clause can't be
+		 * pushed down or needs re-checking.
+		 */
+		if (!clause_context.can_pushdown || clause_context.needs_recheck || chunk_partial)
 		{
 			decompress_clauses = lappend(decompress_clauses, ri);
 		}
@@ -121,21 +132,6 @@ make_segment_meta_opexpr(QualPushdownContext *context, Oid opno, AttrNumber meta
 									uncompressed_var->varcollid);
 }
 
-static Expr *
-get_pushdownsafe_expr(const QualPushdownContext *input_context, Expr *input)
-{
-	/* do not mess up the input_context, so create a new one */
-	QualPushdownContext test_context;
-	Expr *expr;
-
-	memcpy(&test_context, input_context, sizeof(test_context));
-	test_context.can_pushdown = true;
-	expr = (Expr *) modify_expression((Node *) input, &test_context);
-	if (test_context.can_pushdown)
-		return expr;
-	return NULL;
-}
-
 static void
 expr_fetch_minmax_metadata(QualPushdownContext *context, Expr *expr, AttrNumber *min_attno,
 						   AttrNumber *max_attno)
@@ -171,78 +167,92 @@ expr_fetch_minmax_metadata(QualPushdownContext *context, Expr *expr, AttrNumber
 												  "max");
 }
 
-static Expr *
-pushdown_op_to_segment_meta_min_max(QualPushdownContext *context, List *expr_args, Oid op_oid,
-									Oid op_collation)
+static void *
+pushdown_op_to_segment_meta_min_max(QualPushdownContext *context, OpExpr *orig_opexpr)
 {
-	Expr *leftop, *rightop;
-	TypeCacheEntry *tce;
-	int strategy;
-	Oid expr_type_id;
+	/*
+	 * This always requires rechecking the decompressed data.
+	 */
+	context->needs_recheck = true;
 
-	if (list_length(expr_args) != 2)
-		return NULL;
+	List *expr_args = orig_opexpr->args;
+	Assert(list_length(expr_args) == 2);
+	Expr *orig_leftop = linitial(expr_args);
+	Expr *orig_rightop = lsecond(expr_args);
 
-	leftop = linitial(expr_args);
-	rightop = lsecond(expr_args);
-
-	if (IsA(leftop, RelabelType))
-		leftop = ((RelabelType *) leftop)->arg;
-	if (IsA(rightop, RelabelType))
-		rightop = ((RelabelType *) rightop)->arg;
+	if (IsA(orig_leftop, RelabelType))
+		orig_leftop = ((RelabelType *) orig_leftop)->arg;
+	if (IsA(orig_rightop, RelabelType))
+		orig_rightop = ((RelabelType *) orig_rightop)->arg;
 
 	/* Find the side that has var with segment meta set expr to the other side */
+	Oid op_oid = orig_opexpr->opno;
 	AttrNumber min_attno;
 	AttrNumber max_attno;
-	expr_fetch_minmax_metadata(context, leftop, &min_attno, &max_attno);
+	expr_fetch_minmax_metadata(context, orig_leftop, &min_attno, &max_attno);
 	if (min_attno == InvalidAttrNumber || max_attno == InvalidAttrNumber)
 	{
 		/* No metadata for the left operand, try to commute the operator. */
 		op_oid = get_commutator(op_oid);
-		Expr *tmp = leftop;
-		leftop = rightop;
-		rightop = tmp;
+		Expr *tmp = orig_leftop;
+		orig_leftop = orig_rightop;
+		orig_rightop = tmp;
 
-		expr_fetch_minmax_metadata(context, leftop, &min_attno, &max_attno);
+		expr_fetch_minmax_metadata(context, orig_leftop, &min_attno, &max_attno);
 	}
 
 	if (min_attno == InvalidAttrNumber || max_attno == InvalidAttrNumber)
 	{
 		/* No metadata for either operand. */
-		return NULL;
+		context->can_pushdown = false;
+		return orig_opexpr;
 	}
 
-	Var *var_with_segment_meta = castNode(Var, leftop);
-	Expr *expr = rightop;
+	Var *var_with_segment_meta = castNode(Var, orig_leftop);
 
 	/* May be able to allow non-strict operations as well.
 	 * Next steps: Think through edge cases, either allow and write tests or figure out why we must
 	 * block strict operations
 	 */
 	if (!OidIsValid(op_oid) || !op_strict(op_oid))
-		return NULL;
+	{
+		context->can_pushdown = false;
+		return orig_opexpr;
+	}
 
 	/* If the collation to be used by the OP doesn't match the column's collation do not push down
 	 * as the materialized min/max value do not match the semantics of what we need here */
+	Oid op_collation = orig_opexpr->inputcollid;
 	if (var_with_segment_meta->varcollid != op_collation)
-		return NULL;
-
-	tce = lookup_type_cache(var_with_segment_meta->vartype, TYPECACHE_BTREE_OPFAMILY);
-
-	strategy = get_op_opfamily_strategy(op_oid, tce->btree_opf);
-	if (strategy == InvalidStrategy)
-		return NULL;
-
-	Expr *new_expr = get_pushdownsafe_expr(context, expr);
-
-	if (new_expr == NULL)
 	{
-		return NULL;
+		context->can_pushdown = false;
+		return orig_opexpr;
 	}
 
-	expr = new_expr;
+	TypeCacheEntry *tce =
+		lookup_type_cache(var_with_segment_meta->vartype, TYPECACHE_BTREE_OPFAMILY);
 
-	expr_type_id = exprType((Node *) expr);
+	const int strategy = get_op_opfamily_strategy(op_oid, tce->btree_opf);
+	if (strategy == InvalidStrategy)
+	{
+		context->can_pushdown = false;
+		return orig_opexpr;
+	}
+
+	/*
+	 * Check if the righthand expression is safe to push down. We cannot combine
+	 * it with the original operator if there can be false negatives.
+	 */
+	QualPushdownContext tmp_context = copy_context(context);
+	Expr *pushed_down_rightop = (Expr *) qual_pushdown_mutator((Node *) orig_rightop, &tmp_context);
+	if (!tmp_context.can_pushdown || tmp_context.needs_recheck)
+	{
+		context->can_pushdown = false;
+		return orig_opexpr;
+	}
+	Assert(pushed_down_rightop != NULL);
+
+	const Oid expr_type_id = exprType((Node *) pushed_down_rightop);
 
 	switch (strategy)
 	{
@@ -259,20 +269,27 @@ pushdown_op_to_segment_meta_min_max(QualPushdownContext *context, List *expr_arg
 											  BTGreaterEqualStrategyNumber);
 
 			if (!OidIsValid(opno_le) || !OidIsValid(opno_ge))
-				return NULL;
+			{
+				/*
+				 * Shouldn't be possible if we managed to create the min/max
+				 * sparse index, but defend against catalog corruption.
+				 */
+				context->can_pushdown = false;
+				return orig_opexpr;
+			}
 
 			return make_andclause(
 				list_make2(make_segment_meta_opexpr(context,
 													opno_le,
 													min_attno,
 													var_with_segment_meta,
-													expr,
+													pushed_down_rightop,
 													BTLessEqualStrategyNumber),
 						   make_segment_meta_opexpr(context,
 													opno_ge,
 													max_attno,
 													var_with_segment_meta,
-													expr,
+													pushed_down_rightop,
 													BTGreaterEqualStrategyNumber)));
 		}
 		case BTLessStrategyNumber:
@@ -283,13 +300,20 @@ pushdown_op_to_segment_meta_min_max(QualPushdownContext *context, List *expr_arg
 					get_opfamily_member(tce->btree_opf, tce->type_id, expr_type_id, strategy);
 
 				if (!OidIsValid(opno))
-					return NULL;
+				{
+					/*
+					 * Shouldn't be possible if we managed to create the min/max
+					 * sparse index, but defend against catalog corruption.
+					 */
+					context->can_pushdown = false;
+					return orig_opexpr;
+				}
 
 				return (Expr *) make_segment_meta_opexpr(context,
 														 opno,
 														 min_attno,
 														 var_with_segment_meta,
-														 expr,
+														 pushed_down_rightop,
 														 strategy);
 			}
 
@@ -301,17 +325,25 @@ pushdown_op_to_segment_meta_min_max(QualPushdownContext *context, List *expr_arg
 					get_opfamily_member(tce->btree_opf, tce->type_id, expr_type_id, strategy);
 
 				if (!OidIsValid(opno))
-					return NULL;
+				{
+					/*
+					 * Shouldn't be possible if we managed to create the min/max
+					 * sparse index, but defend against catalog corruption.
+					 */
+					context->can_pushdown = false;
+					return orig_opexpr;
+				}
 
 				return (Expr *) make_segment_meta_opexpr(context,
 														 opno,
 														 max_attno,
 														 var_with_segment_meta,
-														 expr,
+														 pushed_down_rightop,
 														 strategy);
 			}
 		default:
-			return NULL;
+			context->can_pushdown = false;
+			return orig_opexpr;
 	}
 }
 
@@ -343,55 +375,57 @@ expr_fetch_bloom1_metadata(QualPushdownContext *context, Expr *expr, AttrNumber
 													 "bloom1");
 }
 
-static Expr *
-pushdown_op_to_segment_meta_bloom1(QualPushdownContext *context, List *expr_args, Oid op_oid,
-								   Oid op_collation)
+static void *
+pushdown_op_to_segment_meta_bloom1(QualPushdownContext *context, OpExpr *orig_opexpr)
 {
-	Expr *original_leftop;
-	Expr *original_rightop;
-	TypeCacheEntry *tce;
-	int strategy;
+	/*
+	 * This always requires rechecking the decompressed data.
+	 */
+	context->needs_recheck = true;
 
-	if (list_length(expr_args) != 2)
-		return NULL;
+	List *expr_args = orig_opexpr->args;
+	Assert(list_length(expr_args) == 2);
+	Expr *orig_leftop = linitial(expr_args);
+	Expr *orig_rightop = lsecond(expr_args);
 
-	original_leftop = linitial(expr_args);
-	original_rightop = lsecond(expr_args);
-
-	if (IsA(original_leftop, RelabelType))
-		original_leftop = ((RelabelType *) original_leftop)->arg;
-	if (IsA(original_rightop, RelabelType))
-		original_rightop = ((RelabelType *) original_rightop)->arg;
+	if (IsA(orig_leftop, RelabelType))
+		orig_leftop = ((RelabelType *) orig_leftop)->arg;
+	if (IsA(orig_rightop, RelabelType))
+		orig_rightop = ((RelabelType *) orig_rightop)->arg;
 
 	/* Find the side that has var with segment meta set expr to the other side */
+	Oid op_oid = orig_opexpr->opno;
 	AttrNumber bloom1_attno = InvalidAttrNumber;
-	expr_fetch_bloom1_metadata(context, original_leftop, &bloom1_attno);
+	expr_fetch_bloom1_metadata(context, orig_leftop, &bloom1_attno);
 	if (bloom1_attno == InvalidAttrNumber)
 	{
 		/* No metadata for the left operand, try to commute the operator. */
 		op_oid = get_commutator(op_oid);
-		Expr *tmp = original_leftop;
-		original_leftop = original_rightop;
-		original_rightop = tmp;
+		Expr *tmp = orig_leftop;
+		orig_leftop = orig_rightop;
+		orig_rightop = tmp;
 
-		expr_fetch_bloom1_metadata(context, original_leftop, &bloom1_attno);
+		expr_fetch_bloom1_metadata(context, orig_leftop, &bloom1_attno);
 	}
 
 	if (bloom1_attno == InvalidAttrNumber)
 	{
 		/* No metadata for either operand. */
-		return NULL;
+		context->can_pushdown = false;
+		return orig_opexpr;
 	}
 
-	Var *var_with_segment_meta = castNode(Var, original_leftop);
+	Var *var_with_segment_meta = castNode(Var, orig_leftop);
 
 	/*
 	 * Play it safe and don't push down if the operator collation doesn't match
 	 * the column collation.
 	 */
+	Oid op_collation = orig_opexpr->inputcollid;
 	if (var_with_segment_meta->varcollid != op_collation)
 	{
-		return NULL;
+		context->can_pushdown = false;
+		return orig_opexpr;
 	}
 
 	/*
@@ -399,17 +433,20 @@ pushdown_op_to_segment_meta_bloom1(QualPushdownContext *context, List *expr_args
 	 */
 	if (OidIsValid(op_collation) && !get_collation_isdeterministic(op_collation))
 	{
-		return NULL;
+		context->can_pushdown = false;
+		return orig_opexpr;
 	}
 
 	/*
 	 * We only support hashable equality operators.
 	 */
-	tce = lookup_type_cache(var_with_segment_meta->vartype, TYPECACHE_HASH_OPFAMILY);
-	strategy = get_op_opfamily_strategy(op_oid, tce->hash_opf);
+	TypeCacheEntry *tce =
+		lookup_type_cache(var_with_segment_meta->vartype, TYPECACHE_HASH_OPFAMILY);
+	const int strategy = get_op_opfamily_strategy(op_oid, tce->hash_opf);
 	if (strategy != HTEqualStrategyNumber)
 	{
-		return NULL;
+		context->can_pushdown = false;
+		return orig_opexpr;
 	}
 
 	/*
@@ -418,14 +455,17 @@ pushdown_op_to_segment_meta_bloom1(QualPushdownContext *context, List *expr_args
 	Assert(op_strict(op_oid));
 
 	/*
-	 * Check if the righthand expression is safe to push down.
-	 * functions.
+	 * Check if the righthand expression is safe to push down. We cannot combine
+	 * it with the original operator if there can be false negatives.
 	 */
-	Expr *pushed_down_rightop = get_pushdownsafe_expr(context, original_rightop);
-	if (pushed_down_rightop == NULL)
+	QualPushdownContext tmp_context = copy_context(context);
+	Expr *pushed_down_rightop = (Expr *) qual_pushdown_mutator((Node *) orig_rightop, &tmp_context);
+	if (!tmp_context.can_pushdown || tmp_context.needs_recheck)
 	{
-		return NULL;
+		context->can_pushdown = false;
+		return orig_opexpr;
 	}
+	Assert(pushed_down_rightop != NULL);
 
 	/*
 	 * var = expr implies bloom1_contains(var_bloom, expr).
@@ -451,86 +491,62 @@ pushdown_op_to_segment_meta_bloom1(QualPushdownContext *context, List *expr_args
 								 COERCE_EXPLICIT_CALL);
 }
 
-static Node *
-modify_expression(Node *node, QualPushdownContext *context)
+static bool
+contain_volatile_functions_checker(Oid func_id, void *context)
 {
-	if (node == NULL)
-		return NULL;
+	return (func_volatile(func_id) == PROVOLATILE_VOLATILE);
+}
 
-	switch (nodeTag(node))
+/*
+ * Push down the given expression node.
+ *
+ * This is used as a mutator for expression_tree_mutator().
+ *
+ * We return the original node if we cannot push it down, to be consistent with
+ * the expression_tree_mutator behavior. The caller must check
+ * context.can_pushdown.
+ */
+static Node *
+qual_pushdown_mutator(Node *orig_node, QualPushdownContext *context)
+{
+	if (orig_node == NULL)
 	{
-		case T_OpExpr:
-		{
-			OpExpr *opexpr = (OpExpr *) node;
-			if (opexpr->opresulttype == BOOLOID)
-			{
-				Expr *pd = NULL;
+		/*
+		 * An expression node can have a NULL field and the mutator will be
+		 * still called for it, so we have to handle this.
+		 */
+		return NULL;
+	}
 
-				if (ts_guc_enable_sparse_index_bloom)
-				{
-					/*
-					 * Try bloom1 sparse index.
-					 */
-					pd = pushdown_op_to_segment_meta_bloom1(context,
-															opexpr->args,
-															opexpr->opno,
+	if (!context->can_pushdown)
+	{
+		/*
+		 * Stop early if we already know we can't push down this filter.
+		 */
+		return orig_node;
+	}
 
-															opexpr->inputcollid);
-				}
+	if (check_functions_in_node(orig_node,
+								contain_volatile_functions_checker,
+								/* context = */ NULL))
+	{
+		/* pushdown is not safe for volatile expressions */
+		context->can_pushdown = false;
+		return orig_node;
+	}
 
-				if (pd != NULL)
-				{
-					context->needs_recheck = true;
-					/* pd is on the compressed table so do not mutate further */
-					return (Node *) pd;
-				}
-
-				/*
-				 * Try minmax sparse index.
-				 */
-				pd = pushdown_op_to_segment_meta_min_max(context,
-														 opexpr->args,
-														 opexpr->opno,
-														 opexpr->inputcollid);
-				if (pd != NULL)
-				{
-					context->needs_recheck = true;
-					/* pd is on the compressed table so do not mutate further */
-					return (Node *) pd;
-				}
-			}
-			/* opexpr will still be checked for segment by columns */
-			break;
-		}
-		case T_BoolExpr:
-		case T_CoerceViaIO:
-		case T_RelabelType:
-		case T_ScalarArrayOpExpr:
-		case T_List:
-		case T_Const:
-		case T_NullTest:
-		case T_Param:
-		case T_SQLValueFunction:
-		case T_CaseExpr:
-		case T_CaseWhen:
-			break;
-		case T_FuncExpr:
-			/*
-			 * The caller should have checked that we don't have volatile
-			 * functions in this qual.
-			 */
-			Assert(!contain_volatile_functions(node));
-			break;
+	switch (nodeTag(orig_node))
+	{
 		case T_Var:
 		{
-			Var *var = castNode(Var, node);
+			Var *var = castNode(Var, orig_node);
 			Assert((Index) var->varno == context->chunk_rel->relid);
 
 			if (var->varattno <= 0)
 			{
 				/* Can't do this for system columns such as whole-row var. */
 				context->can_pushdown = false;
-				return NULL;
+				return orig_node;
 			}
 
 			char *attname = get_attname(context->chunk_rte->relid, var->varattno, false);
@@ -538,7 +554,7 @@ modify_expression(Node *node, QualPushdownContext *context)
 			if (!ts_array_is_member(context->settings->fd.segmentby, attname))
 			{
 				context->can_pushdown = false;
-				return NULL;
+				return orig_node;
 			}
 
 			var = copyObject(var);
@@ -547,11 +563,131 @@ modify_expression(Node *node, QualPushdownContext *context)
 
 			return (Node *) var;
 		}
+		case T_OpExpr:
+		{
+			OpExpr *opexpr = (OpExpr *) orig_node;
+
+			/*
+			 * It might be possible to push down the OpExpr as is, if it
+			 * references only the segmentby columns. Check this case first.
+			 *
+			 * Note that we can't push down the entire operator if we pushed
+			 * down both sides inexactly, i.e. they require recheck. This means
+			 * we can have false positives there, and combining false positives
+			 * with the original operator could lead to false negatives, which
+			 * would be a bug. Consider for example (x = 1) = (y = 1) in case
+			 * where both sides are false, but there's a false posistive for the
+			 * pushed down version of the left side but not the right side.
+			 */
+			QualPushdownContext tmp_context = copy_context(context);
+			void *pushed_down =
+				expression_tree_mutator((Node *) orig_node, qual_pushdown_mutator, &tmp_context);
+			if (tmp_context.can_pushdown && !tmp_context.needs_recheck)
+			{
+				return pushed_down;
+			}
+
+			if (opexpr->opresulttype != BOOLOID)
+			{
+				/*
+				 * The following pushdown options only support operators that
+				 * return bool.
+				 */
+				context->can_pushdown = false;
+				return orig_node;
+			}
+
+			if (list_length(opexpr->args) != 2)
+			{
+				/*
+				 * The following pushdown options only support operators with
+				 * two operands.
+				 */
+				context->can_pushdown = false;
+				return orig_node;
+			}
+
+			/*
+			 * Try bloom1 sparse index.
+			 */
+			if (ts_guc_enable_sparse_index_bloom)
+			{
+				tmp_context = copy_context(context);
+				pushed_down = pushdown_op_to_segment_meta_bloom1(&tmp_context, opexpr);
+				if (tmp_context.can_pushdown)
+				{
+					context->needs_recheck |= tmp_context.needs_recheck;
+					return pushed_down;
+				}
+			}
+
+			/*
+			 * Try minmax sparse index.
+			 */
+			tmp_context = copy_context(context);
+			pushed_down = pushdown_op_to_segment_meta_min_max(&tmp_context, opexpr);
+			if (tmp_context.can_pushdown)
+			{
+				context->needs_recheck |= tmp_context.needs_recheck;
+				return pushed_down;
+			}
+
+			/*
+			 * No other options to push down the OpExpr.
+			 */
+			context->can_pushdown = false;
+			return orig_node;
+		}
+		case T_ScalarArrayOpExpr:
+		{
+			/*
+			 * It can be possible to push down the scalar array operation as is,
+			 * if it references only the segmentby columns. Check this case
+			 * first.
+			 *
+			 * See the comment for OpExpr about needs_recheck handling.
+			 */
+			QualPushdownContext tmp_context = copy_context(context);
+			void *pushed_down =
+				expression_tree_mutator((Node *) orig_node, qual_pushdown_mutator, &tmp_context);
+			if (tmp_context.can_pushdown && !tmp_context.needs_recheck)
+			{
+				return pushed_down;
+			}
+
+			/*
+			 * No other ways to push it down, so consider it failed.
+			 */
+			context->can_pushdown = false;
+			return orig_node;
+		}
+		/*
+		 * These nodes do not influence the pushdown by themselves, so we
+		 * recurse.
+		 */
+		case T_BoolExpr:
+		case T_FuncExpr:
+		case T_CoerceViaIO:
+		case T_RelabelType:
+		case T_List:
+		case T_Const:
+		case T_NullTest:
+		case T_Param:
+		case T_SQLValueFunction:
+		case T_CaseExpr:
+		case T_CaseWhen:
+		case T_ArrayExpr:
+		{
+			Node *pushed_down =
+				expression_tree_mutator((Node *) orig_node, qual_pushdown_mutator, context);
+			return pushed_down;
+		}
+
+		/*
+		 * We don't know how to work with other nodes.
+		 */
 		default:
 			context->can_pushdown = false;
-			return NULL;
-			break;
+			return orig_node;
 	}
-
-	return expression_tree_mutator(node, modify_expression, context);
 }
