@@ -215,4 +215,89 @@ BEGIN;
 CREATE TABLE t24(time timestamptz NOT NULL, device text, value float) WITH (tsdb.hypertable,tsdb.partition_column='time',tsdb.enable_columnstore=false);
 SELECT hypertable_name, compression_enabled FROM timescaledb_information.hypertables;
 ROLLBACK;
+BEGIN;
+CREATE TABLE t25(time timestamptz NOT NULL, device text, value float) WITH (tsdb.hypertable,tsdb.partition_column='time');
+SELECT * FROM _timescaledb_catalog.compression_settings WHERE relid = 't25'::regclass;
+ROLLBACK;
+BEGIN;
+-- don't allow empty orderby
+\set ON_ERROR_STOP 0
+CREATE TABLE t26(time timestamptz NOT NULL, device text, value float) WITH (tsdb.hypertable,tsdb.partition_column='time', tsdb.orderby = '');
+SELECT * FROM _timescaledb_catalog.compression_settings WHERE relid = 't26'::regclass;
+ROLLBACK;
+\set ON_ERROR_STOP 1
+-- can allow empty segmentby
+BEGIN;
+CREATE TABLE t27(time timestamptz NOT NULL, device text, value float) WITH (tsdb.hypertable,tsdb.partition_column='time', tsdb.segmentby = '');
+SELECT * FROM _timescaledb_catalog.compression_settings WHERE relid = 't27'::regclass;
+ROLLBACK;
 
+-- test configurable bloom and minmax
+BEGIN;
+CREATE TABLE t28(time timestamptz NOT NULL, device text, value float) WITH (tsdb.hypertable,tsdb.partition_column='time', tsdb.orderby='time', tsdb.index='bloom(value)');
+SELECT * FROM _timescaledb_catalog.compression_settings;
+ROLLBACK;
+
+BEGIN;
+CREATE TABLE t29(time timestamptz NOT NULL, device text, value float) WITH (tsdb.hypertable,tsdb.partition_column='time', tsdb.orderby='time', tsdb.index='bloom(value),minmax(device)');
+SELECT * FROM _timescaledb_catalog.compression_settings;
+ROLLBACK;
+
+--multi column
+BEGIN;
+CREATE TABLE t30(time timestamptz NOT NULL, device text, value float) WITH (tsdb.hypertable,tsdb.partition_column='time', tsdb.orderby='time',tsdb.index='bloom(value),bloom(device)');
+SELECT * FROM _timescaledb_catalog.compression_settings;
+ROLLBACK;
+
+--test errors
+\set ON_ERROR_STOP 0
+-- invalid syntax
+BEGIN;
+CREATE TABLE t31(time timestamptz NOT NULL, device text, value float) WITH (tsdb.hypertable,tsdb.partition_column='time', tsdb.orderby='value', tsdb.index='bloom(value),bloom(count(device))');
+ROLLBACK;
+
+BEGIN;
+CREATE TABLE t32(time timestamptz NOT NULL, device text, value float) WITH (tsdb.hypertable,tsdb.partition_column='time', tsdb.orderby='value', tsdb.index='value,bloom(count(device))');
+ROLLBACK;
+
+-- duplicate column
+BEGIN;
+CREATE TABLE t33(time timestamptz NOT NULL, device text, value float) WITH (tsdb.hypertable,tsdb.partition_column='time', tsdb.orderby='value', tsdb.index='bloom(value),minmax(value)');
+ROLLBACK;
+
+BEGIN;
+CREATE TABLE t34(time timestamptz NOT NULL, device text, value float) WITH (tsdb.hypertable,tsdb.partition_column='time', tsdb.orderby='value', tsdb.index='bloom(value),bloom(value)');
+ROLLBACK;
+
+-- invalid column
+BEGIN;
+CREATE TABLE t35(time timestamptz NOT NULL, device text, value float) WITH (tsdb.hypertable,tsdb.partition_column='time', tsdb.orderby='value', tsdb.index='bloom(value),bloom(foo)');
+ROLLBACK;
+
+-- same column as orderby
+BEGIN;
+CREATE TABLE t36(time timestamptz NOT NULL, device text, value float) WITH (tsdb.hypertable,tsdb.partition_column='time', tsdb.orderby='value', tsdb.index='bloom(value)');
+ROLLBACK;
+
+BEGIN;
+CREATE TABLE t37(time timestamptz NOT NULL, device text, value float) WITH (tsdb.hypertable,tsdb.partition_column='time', tsdb.orderby='value', tsdb.index='minmax(value)');
+ROLLBACK;
+
+-- same column as segmentby
+BEGIN;
+CREATE TABLE t38(time timestamptz NOT NULL, device text, value float) WITH (tsdb.hypertable,tsdb.partition_column='time', tsdb.orderby='value', tsdb.segmentby='device', tsdb.index='minmax(device)');
+ROLLBACK;
+
+-- guc disabled
+set timescaledb.enable_sparse_index_bloom to false;
+
+BEGIN;
+CREATE TABLE t39(time timestamptz NOT NULL, device text, value float) WITH (tsdb.hypertable,tsdb.partition_column='time', tsdb.orderby='value', tsdb.index='bloom(value)');
+ROLLBACK;
+
+--no orderby
+BEGIN;
+CREATE TABLE t40(time timestamptz NOT NULL, device text, value float) WITH (tsdb.hypertable,tsdb.partition_column='time', tsdb.index='bloom(value)');
+ROLLBACK;
+reset timescaledb.enable_sparse_index_bloom;
+\set ON_ERROR_STOP 1
\ No newline at end of file
